"""
Unfortunately right now Robotframework doesn't really provide the needed hooks
for a debugger, so, we monkey-patch internal APIs to gather the needed info.

More specifically:

    robot.running.steprunner.StepRunner - def run_step

    is patched so that we can stop when some line is about to be executed.
"""
import functools
from robotframework_debug_adapter import file_utils
import threading
from robotframework_debug_adapter.constants import (
    STATE_RUNNING,
    STATE_PAUSED,
    ReasonEnum,
    StepEnum,
)
import itertools
from functools import partial, lru_cache
import os.path
from robocorp_ls_core.robotframework_log import get_logger, get_log_level
from collections import namedtuple
import weakref
from robotframework_debug_adapter.protocols import (
    IRobotDebugger,
    INextId,
    IRobotBreakpoint,
    IBusyWait,
    IEvaluationInfo,
)
from typing import Optional, List, Iterable, Union, Any, Dict, FrozenSet
from robocorp_ls_core.basic import implements
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    StackFrame,
    Scope,
    Source,
    Variable,
    OutputEvent,
    OutputEventBody,
)

from robotframework_ls.impl.robot_constants import (
    get_builtin_variables,
    ROBOT_AND_TXT_FILE_EXTENSIONS,
)


@lru_cache(None)
def get_builtin_normalized_names() -> FrozenSet[str]:
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    normalized = list()
    for k, _ in get_builtin_variables():
        normalized.append(normalize_robot_name(k))
    return frozenset(normalized)


log = get_logger(__name__)

next_id: INextId = partial(next, itertools.count(1))


class RobotBreakpoint(object):
    def __init__(
        self,
        lineno: int,
        condition: Optional[str] = None,
        hit_condition: Optional[int] = None,
        log_message: Optional[str] = None,
    ):
        """
        :param lineno:
            1-based line for the breakpoint.

        :param condition:
            If specified, the breakpoint will only be hit if the condition
            evaluates to True.

        :param hit_condition:
            If specified, the breakpoint will only be hit after it hits the
            specified number of times.

        :param log_message:
            If specified, the breakpoint will not actually break, it'll just
            print the given message instead of breaking.
        """
        self.lineno = lineno
        self.condition = condition
        if hit_condition is not None:
            hit_condition = int(hit_condition)
        self.hit_condition = hit_condition
        self.log_message = log_message
        self.hits = 0


class BusyWait(object):
    def __init__(self):
        self.before_wait = []
        self.waited = 0
        self.proceeded = 0
        self._condition = threading.Condition()

    @implements(IBusyWait.pre_wait)
    def pre_wait(self):
        for c in self.before_wait:
            c()

    @implements(IBusyWait.wait)
    def wait(self):
        self.waited += 1
        with self._condition:
            self._condition.wait()

    @implements(IBusyWait.proceed)
    def proceed(self):
        self.proceeded += 1
        with self._condition:
            self._condition.notify_all()


class _BaseObjectToDAP(object):
    """
    Base class for classes which converts some object to the DAP.
    """

    def compute_as_dap(self) -> List[Variable]:
        return []


class _ArgsAsDAP(_BaseObjectToDAP):
    """
    Provides args as DAP variables.
    """

    def __init__(self, keyword_args):
        self._keyword_args = keyword_args

    def compute_as_dap(self) -> List[Variable]:
        from robotframework_debug_adapter.safe_repr import SafeRepr

        lst = []
        safe_repr = SafeRepr()
        for i, arg in enumerate(self._keyword_args):
            lst.append(Variable("Arg %s" % (i,), safe_repr(arg), variablesReference=0))
        return lst


class _NonBuiltinVariablesAsDAP(_BaseObjectToDAP):
    """
    Provides variables as DAP variables.
    """

    def __init__(self, variables):
        self._variables = variables
        self._builtins = get_builtin_normalized_names()

    def compute_as_dap(self) -> List[Variable]:
        from robotframework_debug_adapter.safe_repr import SafeRepr

        variables = self._variables
        as_dct = variables.as_dict()
        lst = []
        safe_repr = SafeRepr()

        for key, val in as_dct.items():
            if self._accept(key):
                lst.append(
                    Variable(safe_repr(key), safe_repr(val), variablesReference=0)
                )
        return lst

    def _accept(self, k: str) -> bool:
        from robotframework_ls.impl.variable_resolve import normalize_variable_name

        if normalize_variable_name(k) in self._builtins:
            return False
        else:
            return True


class _BuiltinsAsDAP(_NonBuiltinVariablesAsDAP):
    """
    Provides variables as DAP variables.
    """

    def _accept(self, k: str) -> bool:
        return not _NonBuiltinVariablesAsDAP._accept(self, k)


class _BaseFrameInfo(object):
    @property
    def dap_frame(self):
        raise NotImplementedError("Not implemented in: %s" % (self.__class__,))

    def get_scopes(self) -> List[Scope]:
        raise NotImplementedError("Not implemented in: %s" % (self.__class__,))

    def get_type_name(self):
        raise NotImplementedError("Not implemented in: %s" % (self.__class__,))


class _SuiteFrameInfo(_BaseFrameInfo):
    def __init__(self, stack_list, dap_frame):
        self._stack_list = weakref.ref(stack_list)
        self._dap_frame = dap_frame

    @property
    def dap_frame(self):
        return self._dap_frame

    def get_scopes(self) -> List[Scope]:
        return []

    def get_type_name(self):
        return "Suite"


class _TestFrameInfo(_BaseFrameInfo):
    def __init__(self, stack_list, dap_frame):
        self._stack_list = weakref.ref(stack_list)
        self._dap_frame = dap_frame

    @property
    def dap_frame(self):
        return self._dap_frame

    def get_scopes(self) -> List[Scope]:
        return []

    def get_type_name(self):
        return "Test"


class _KeywordFrameInfo(_BaseFrameInfo):
    def __init__(
        self, stack_list, dap_frame, name, lineno, args, variables, execution_context
    ):
        self._stack_list = weakref.ref(stack_list)
        self._dap_frame = dap_frame
        self._name = name
        self._lineno = lineno
        self._scopes = None
        self._args = args
        self._variables = variables
        self._execution_context = execution_context

    @property
    def name(self):
        return self._name

    @property
    def lineno(self):
        return self._lineno

    @property
    def variables(self):
        return self._variables

    @property
    def execution_context(self):
        return self._execution_context

    @property
    def dap_frame(self):
        return self._dap_frame

    def get_type_name(self):
        return "Keyword"

    def get_scopes(self) -> List[Scope]:
        if self._scopes is not None:
            return self._scopes
        stack_list = self._stack_list()
        if stack_list is None:
            return []

        locals_variables_reference: int = next_id()
        vars_variables_reference: int = next_id()
        builtions_variables_reference: int = next_id()
        scopes = [
            Scope("Variables", vars_variables_reference, expensive=False),
            Scope(
                "Arguments",
                locals_variables_reference,
                expensive=False,
                presentationHint="locals",
            ),
            Scope("Builtins", builtions_variables_reference, expensive=False),
        ]

        args = self._args
        stack_list.register_variables_reference(
            locals_variables_reference, _ArgsAsDAP(args)
        )
        # ctx.namespace.get_library_instances()

        stack_list.register_variables_reference(
            vars_variables_reference, _NonBuiltinVariablesAsDAP(self._variables)
        )
        stack_list.register_variables_reference(
            builtions_variables_reference, _BuiltinsAsDAP(self._variables)
        )
        self._scopes = scopes
        return self._scopes


class _StackInfo(object):
    """
    This is the information for the stacks available when we're stopped in a
    breakpoint.
    """

    def __init__(self):
        self._frame_id_to_frame_info: Dict[int, _BaseFrameInfo] = {}
        self._dap_frames = []
        self._ref_id_to_children = {}

    def iter_frame_ids(self) -> Iterable[int]:
        """
        Access to list(int) where iter_frame_ids[0] is the current frame
        where we're stopped (topmost frame).
        """
        return (x.id for x in self._dap_frames)

    def register_variables_reference(self, variables_reference, children):
        self._ref_id_to_children[variables_reference] = children

    def add_keyword_entry_stack(
        self, name, lineno, filename: str, args, variables, execution_context
    ) -> int:
        frame_id: int = next_id()
        dap_frame = StackFrame(
            frame_id,
            name=name,
            line=lineno or 1,
            column=0,
            source=Source(name=os.path.basename(filename), path=filename),
        )
        self._dap_frames.append(dap_frame)
        self._frame_id_to_frame_info[frame_id] = _KeywordFrameInfo(
            self, dap_frame, name, lineno, args, variables, execution_context
        )
        return frame_id

    def add_suite_entry_stack(self, name: str, filename: str) -> int:
        frame_id: int = next_id()
        dap_frame = StackFrame(
            frame_id,
            name=name,
            line=1,
            column=0,
            source=Source(name=os.path.basename(filename), path=filename),
        )
        self._dap_frames.append(dap_frame)
        self._frame_id_to_frame_info[frame_id] = _SuiteFrameInfo(self, dap_frame)
        return frame_id

    def add_test_entry_stack(self, name: str, filename: str, lineno: int) -> int:
        from robocorp_ls_core.debug_adapter_core.dap import dap_schema

        frame_id: int = next_id()
        dap_frame = dap_schema.StackFrame(
            frame_id,
            name=name,
            line=lineno,
            column=0,
            source=dap_schema.Source(name=os.path.basename(filename), path=filename),
        )
        self._dap_frames.append(dap_frame)
        self._frame_id_to_frame_info[frame_id] = _TestFrameInfo(self, dap_frame)
        return frame_id

    @property
    def dap_frames(self) -> List[StackFrame]:
        """
        Access to list(StackFrame) where dap_frames[0] is the current frame
        where we're stopped (topmost frame).
        """
        return self._dap_frames

    def get_scopes(self, frame_id):
        frame_info = self._frame_id_to_frame_info.get(frame_id)
        if frame_info is None:
            return None
        return frame_info.get_scopes()

    def get_variables(self, variables_reference):
        lst = self._ref_id_to_children.get(variables_reference)
        if lst is not None:
            if isinstance(lst, _BaseObjectToDAP):
                lst = lst.compute_as_dap()
        return lst


_StepEntry = namedtuple(
    "_StepEntry", "name, lineno, source, args, variables, entry_type, execution_context"
)
_SuiteEntry = namedtuple("_SuiteEntry", "name, source, entry_type")
_TestEntry = namedtuple("_TestEntry", "name, source, lineno, entry_type")
_LogEntry = namedtuple("_LogEntry", "name, source, lineno, entry_type")


class InvalidFrameIdError(Exception):
    pass


class InvalidFrameTypeError(Exception):
    pass


class UnableToEvaluateError(Exception):
    pass


class EvaluationResult(Exception):
    def __init__(self, result):
        self.result = result


class _EvaluationInfo(object):
    def __init__(self, frame_id: int, expression: str, context: str):
        from concurrent import futures
        from robocorp_ls_core.protocols import IFuture

        self.frame_id = frame_id
        self.expression = expression
        self.context = context
        self.future: IFuture[Any] = futures.Future()

    def _do_eval(self, debugger_impl):
        frame_id = self.frame_id
        stack_info = debugger_impl._get_stack_info_from_frame_id(frame_id)

        if stack_info is None:
            raise InvalidFrameIdError(
                "Unable to find frame id for evaluation: %s" % (frame_id,)
            )

        info = stack_info._frame_id_to_frame_info.get(frame_id)
        if info is None:
            raise InvalidFrameIdError(
                "Unable to find frame info for evaluation: %s" % (frame_id,)
            )

        if not isinstance(info, _KeywordFrameInfo):
            raise InvalidFrameTypeError(
                "Can only evaluate at a Keyword context (current context: %s)"
                % (info.get_type_name(),)
            )
        log.info("Doing evaluation in the Keyword context: %s", info.name)

        from robotframework_ls.impl.text_utilities import is_variable_text

        from robot.libraries.BuiltIn import BuiltIn  # type: ignore
        from robot.api import get_model  # type: ignore
        from robotframework_ls.impl import ast_utils

        # We can't really use
        # BuiltIn().evaluate(expression, modules, namespace)
        # because we can't set the variable_store used with it
        # (it always uses the latest).

        variable_store = info.variables.store

        if is_variable_text(self.expression):
            try:
                value = variable_store[self.expression[2:-1]]
            except Exception:
                pass
            else:
                return EvaluationResult(value)

        if self.context == "hover":
            try:
                ctx = info.execution_context
                return EvaluationResult(
                    ctx.namespace.get_runner(self.expression).longname
                )
            except:
                log.exception("Error on hover evaluation: %s", self.expression)
                return EvaluationResult("")

        # Do we want this?
        # from robot.variables.evaluation import evaluate_expression
        # try:
        #     result = evaluate_expression(self.expression, variable_store)
        # except Exception:
        #     log.exception()
        # else:
        #     return EvaluationResult(result)

        # Try to check if it's a KeywordCall.
        s = """
*** Test Cases ***
Evaluation
    %s
""" % (
            self.expression,
        )
        model = get_model(s)
        usage_info = list(
            ast_utils.iter_keyword_usage_tokens(model, collect_args_as_keywords=False)
        )
        if len(usage_info) == 1:
            usage = next(iter(usage_info))
            node = usage.node
            name = usage.name

            dap_frames = stack_info.dap_frames
            if dap_frames:
                top_frame_id = dap_frames[0].id
                if top_frame_id != frame_id:
                    if get_log_level() >= 2:
                        log.debug(
                            "Unable to evaluate.\nFrame id for evaluation: %r\nTop frame id: %r.\nDAP frames:\n%s",
                            frame_id,
                            top_frame_id,
                            "\n".join(x.to_json() for x in dap_frames),
                        )

                    raise UnableToEvaluateError(
                        "Keyword calls may only be evaluated at the topmost frame."
                    )

                assign = node.assign
                from robot.running import Keyword

                kw = Keyword(name, args=node.args, assign=assign)
                ctx = info.execution_context
                return EvaluationResult(kw.run(ctx))

        raise UnableToEvaluateError("Unable to evaluate: %s" % (self.expression,))

    def evaluate(self, debugger_impl):
        """
        :param _RobotDebuggerImpl debugger_impl:
        """
        try:
            r = self._do_eval(debugger_impl)
            self.future.set_result(r.result)
        except Exception as e:
            if get_log_level() >= 2:
                log.exception("Error evaluating: %s", (self.expression,))
            self.future.set_exception(e)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IEvaluationInfo = check_implements(self)


class _RobotDebuggerImpl(object):
    """
    This class provides the main API to deal with debugging
    Robot Framework.
    """

    def __init__(self):
        self.reset()

    @implements(IRobotDebugger.reset)
    def reset(self):
        from collections import deque
        from robotframework_debug_adapter._ignore_failures_in_stack import (
            IgnoreFailuresInStack,
        )

        self._filename_to_line_to_breakpoint = {}
        self.busy_wait = BusyWait()

        self._run_state = STATE_RUNNING
        self._step_cmd: StepEnum = StepEnum.STEP_NONE
        self._reason: ReasonEnum = ReasonEnum.REASON_NOT_STOPPED
        self._next_id = next_id
        self._stack_ctx_entries_deque = deque()
        self._stop_on_stack_len = 0

        self._tid_to_stack_info: Dict[int, _StackInfo] = {}
        self._frame_id_to_tid = {}
        self._evaluations = []
        self._skip_breakpoints = 0

        self._exc_name = None
        self._exc_description = None

        def is_true_in_env(key):
            return os.getenv(key, "0").lower() in (
                "1",
                "True",
                "true",
            )

        self.break_on_log_failure = is_true_in_env("RFLS_BREAK_ON_FAILURE")
        self.break_on_log_error = is_true_in_env("RFLS_BREAK_ON_ERROR")
        self._ignore_failures_in_stack = IgnoreFailuresInStack()

    def write_message(self, msg):
        log.critical(
            "Message: %s not sent!\nExpected _RobotDebuggerImpl.write_message to be replaced to the actual implementation!",
            msg,
        )
        raise AssertionError("Error")

    @property
    def stop_reason(self) -> ReasonEnum:
        return self._reason

    @property
    def exc_name(self) -> Optional[str]:
        return self._exc_name

    @property
    def exc_description(self) -> Optional[str]:
        return self._exc_description

    def _get_stack_info_from_frame_id(self, frame_id) -> Optional[_StackInfo]:
        thread_id = self._frame_id_to_tid.get(frame_id)
        if thread_id is not None:
            return self._get_stack_info(thread_id)
        return None

    def _get_stack_info(self, thread_id) -> Optional[_StackInfo]:
        return self._tid_to_stack_info.get(thread_id)

    def get_frames(self, thread_id) -> Optional[List[StackFrame]]:
        stack_info = self._get_stack_info(thread_id)
        if not stack_info:
            return None
        return stack_info.dap_frames

    def iter_frame_ids(self, thread_id) -> Iterable[int]:
        stack_info = self._get_stack_info(thread_id)
        if not stack_info:
            return ()
        return stack_info.iter_frame_ids()

    def get_scopes(self, frame_id) -> Optional[List[Scope]]:
        tid = self._frame_id_to_tid.get(frame_id)
        if tid is None:
            return None

        stack_info = self._get_stack_info(tid)
        if not stack_info:
            return None
        return stack_info.get_scopes(frame_id)

    def get_variables(self, variables_reference):
        for stack_list in list(self._tid_to_stack_info.values()):
            variables = stack_list.get_variables(variables_reference)
            if variables is not None:
                return variables
        return None

    def _get_filename(self, obj, msg) -> str:
        try:
            source = obj.source
            if source is None:
                return "None"

            filename, _changed = file_utils.norm_file_to_client(source)
        except:
            filename = "<Unable to get %s filename>" % (msg,)
            log.exception(filename)

        return filename

    def _create_stack_info(self, thread_id: int):
        stack_info = _StackInfo()

        for entry in reversed(self._stack_ctx_entries_deque):
            try:
                if entry.__class__ == _StepEntry:
                    name = entry.name
                    lineno = entry.lineno
                    variables = entry.variables
                    args = entry.args
                    execution_context = entry.execution_context
                    filename = self._get_filename(entry, "Keyword")

                    frame_id = stack_info.add_keyword_entry_stack(
                        name, lineno, filename, args, variables, execution_context
                    )

                elif entry.__class__ == _SuiteEntry:
                    name = "TestSuite: %s" % (entry.name,)
                    filename = self._get_filename(entry, "TestSuite")

                    frame_id = stack_info.add_suite_entry_stack(name, filename)

                elif entry.__class__ == _TestEntry:
                    name = "TestCase: %s" % (entry.name,)
                    filename = self._get_filename(entry, "TestCase")

                    frame_id = stack_info.add_test_entry_stack(
                        name, filename, entry.lineno
                    )

                elif entry.__class__ == _LogEntry:
                    name = "Log (%s)" % (entry.name,)
                    filename = self._get_filename(entry, "Log")

                    frame_id = stack_info.add_test_entry_stack(
                        name, filename, entry.lineno
                    )
            except:
                log.exception("Error creating stack trace.")

        for frame_id in stack_info.iter_frame_ids():
            self._frame_id_to_tid[frame_id] = thread_id

        self._tid_to_stack_info[thread_id] = stack_info

    def _dispose_stack_info(self, thread_id):
        stack_list = self._tid_to_stack_info.pop(thread_id)
        for frame_id in stack_list.iter_frame_ids():
            self._frame_id_to_tid.pop(frame_id)

    def get_current_thread_id(self, thread=None):
        from robotframework_debug_adapter.vendored import force_pydevd  # noqa
        from _pydevd_bundle.pydevd_constants import get_current_thread_id

        if thread is None:
            thread = threading.current_thread()
        return get_current_thread_id(thread)

    def wait_suspended(self, reason: ReasonEnum) -> None:
        thread_id = self.get_current_thread_id()

        if self._exc_name or self._exc_description:
            log.info(
                "wait_suspended. Reason: %s. Exc name: %s. Exc description: %s",
                reason,
                self._exc_name,
                self._exc_description,
            )
        else:
            log.info("wait_suspended. Reason: %s", reason)
        self._create_stack_info(thread_id)
        try:
            self._run_state = STATE_PAUSED
            self._reason = reason

            self.busy_wait.pre_wait()

            while self._run_state == STATE_PAUSED:
                self.busy_wait.wait()

                evaluations = self._evaluations
                self._evaluations = []

                for evaluation in evaluations:  #: :type evaluation: _EvaluationInfo
                    self._skip_breakpoints += 1
                    try:
                        evaluation.evaluate(self)
                    finally:
                        self._skip_breakpoints -= 1

            if self._step_cmd == StepEnum.STEP_NEXT:
                self._stop_on_stack_len = len(self._stack_ctx_entries_deque)
                if self._stop_on_stack_len:
                    if self._is_control_step(
                        self._stack_ctx_entries_deque[-1].entry_type
                    ):
                        self._stop_on_stack_len += 1

            elif self._step_cmd == StepEnum.STEP_OUT:
                self._stop_on_stack_len = len(self._stack_ctx_entries_deque) - 1

        finally:
            self._reason = ReasonEnum.REASON_NOT_STOPPED
            self._dispose_stack_info(thread_id)

    @implements(IRobotDebugger.evaluate)
    def evaluate(
        self, frame_id: int, expression: str, context: str = "watch"
    ) -> IEvaluationInfo:
        """
        Asks something to be evaluated.

        This is an asynchronous operation and returns an _EvaluationInfo (to get
        the result, access _EvaluationInfo.future.result())
        """
        evaluation_info = _EvaluationInfo(frame_id, expression, context)
        self._evaluations.append(evaluation_info)
        self.busy_wait.proceed()
        return evaluation_info

    @implements(IRobotDebugger.step_continue)
    def step_continue(self) -> None:
        self._step_cmd = StepEnum.STEP_NONE
        self._run_state = STATE_RUNNING
        self.busy_wait.proceed()

    @implements(IRobotDebugger.step_in)
    def step_in(self) -> None:
        self._step_cmd = StepEnum.STEP_IN
        self._run_state = STATE_RUNNING
        self.busy_wait.proceed()

    @implements(IRobotDebugger.step_next)
    def step_next(self) -> None:
        self._step_cmd = StepEnum.STEP_NEXT
        self._run_state = STATE_RUNNING
        self.busy_wait.proceed()

    @implements(IRobotDebugger.step_out)
    def step_out(self) -> None:
        self._step_cmd = StepEnum.STEP_OUT
        self._run_state = STATE_RUNNING
        self.busy_wait.proceed()

    @implements(IRobotDebugger.set_breakpoints)
    def set_breakpoints(
        self,
        filename: str,
        breakpoints: Union[IRobotBreakpoint, Iterable[IRobotBreakpoint]],
    ) -> None:
        iter_in: Any
        if isinstance(breakpoints, (list, tuple, set)):
            iter_in = breakpoints
        else:
            iter_in = [breakpoints]
        filename = file_utils.get_abs_path_real_path_and_base_from_file(filename)[1]
        line_to_bp = {}

        for bp in iter_in:
            log.info("Set breakpoint in %s: %s", filename, bp.lineno)
            line_to_bp[bp.lineno] = bp
        self._filename_to_line_to_breakpoint[filename] = line_to_bp

    # ------------------------------------------------- RobotFramework listeners

    # 4.0 versions where the lineno is available on the V2 listener
    def start_keyword_v2(self, _name, attributes):
        from robot.running.context import EXECUTION_CONTEXTS

        ctx = EXECUTION_CONTEXTS.current
        lineno = attributes["lineno"]
        source = attributes["source"]
        name = attributes["kwname"]
        args = attributes["args"]
        entry_type = attributes.get("type", "KEYWORD")
        if attributes.get("status") == "NOT RUN":
            return
        if not args:
            args = []
        self._before_run_step(ctx, name, entry_type, lineno, source, args)

    # 4.0 versions where the lineno is available on the V2 listener
    def end_keyword_v2(self, name, attributes):

        if attributes.get("status") == "NOT RUN":
            return

        self._after_run_step()

    # 3.x versions where the lineno is NOT available on the V2 listener
    def before_control_flow_stmt(self, control_flow_stmt, ctx, *args, **kwargs):
        name = ""
        try:
            if control_flow_stmt.type == "IF/ELSE ROOT":
                name = control_flow_stmt.body[0].condition

            elif control_flow_stmt.type == "KEYWORD":
                name = control_flow_stmt.name

            else:
                name = str(control_flow_stmt).strip()
        except:
            pass

        if not name:
            name = control_flow_stmt.__class__.__name__

        try:
            lineno = control_flow_stmt.lineno
            source = control_flow_stmt.source
        except AttributeError:
            return

        try:
            args = control_flow_stmt.args
        except AttributeError:
            args = []
        entry_type = "KEYWORD"
        self._before_run_step(ctx, name, entry_type, lineno, source, args)

    # 3.x versions where the lineno is NOT available on the V2 listener
    def after_control_flow_stmt(self, control_flow_stmt, ctx, *args, **kwargs):
        self._after_run_step()

    def before_keyword_runner(self, runner, step, *args, **kwargs):
        name = ""
        try:
            name = step.name
        except:
            pass
        if not name:
            name = step.__class__.__name__
        try:
            lineno = step.lineno
            source = step.source
        except AttributeError:
            return
        try:
            args = step.args
        except AttributeError:
            args = []
        ctx = runner._context
        entry_type = "KEYWORD"
        self._before_run_step(ctx, name, entry_type, lineno, source, args)

    def after_keyword_runner(self, runner, step, *args, **kwargs):
        self._after_run_step()

    # 3.x versions where the lineno is NOT available on the V2 listener
    def before_run_step(self, step_runner, step, name=None):
        try:
            name = str(step).strip()
            if not name:
                name = step.__class__.__name__
        except:
            name = "<Unable to get keyword name>"
        try:
            lineno = step.lineno
            source = step.source
        except AttributeError:
            return
        try:
            args = step.args
        except AttributeError:
            args = []
        ctx = step_runner._context
        entry_type = "KEYWORD"
        self._before_run_step(ctx, name, entry_type, lineno, source, args)

    # 3.x versions where the lineno is NOT available on the V2 listener
    def after_run_step(self, step_runner, step, name=None):
        self._after_run_step()

    def _is_control_step(self, entry_type):
        return entry_type in (
            "ELSE IF",
            "ELSE",
            "EXCEPT",
            "FINALLY",
            "FOR ITERATION",
            "FOR",
            "IF",
            "ITERATION",
            "TRY",
            "WHILE",
        )

    def _before_run_step(self, ctx, name, entry_type, lineno, source, args):
        if entry_type == "KEYWORD":
            self._ignore_failures_in_stack.push(name)

        if not name:
            name = entry_type

        if self._is_control_step(entry_type):
            self._stop_on_stack_len += 1

        if source is None or lineno is None:
            # RunKeywordIf doesn't have a source, so, just show the caller source.
            for entry in reversed(self._stack_ctx_entries_deque):
                if source is None:
                    source = entry.source
                if lineno is None:
                    lineno = entry.lineno
                break

        if not source:
            return
        if not source.endswith(ROBOT_AND_TXT_FILE_EXTENSIONS):
            robot_init = os.path.join(source, "__init__.robot")
            if os.path.exists(robot_init):
                source = robot_init
        self._stack_ctx_entries_deque.append(
            _StepEntry(
                name, lineno, source, args, ctx.variables.current, entry_type, ctx
            )
        )
        if self._skip_breakpoints:
            return

        source = file_utils.get_abs_path_real_path_and_base_from_file(source)[1]
        log.debug(
            "run_step %s, %s - step: %s - %s\n", name, lineno, self._step_cmd, source
        )
        lines = self._filename_to_line_to_breakpoint.get(source)

        stop_reason: Optional[ReasonEnum] = None
        step_cmd = self._step_cmd
        if lines:
            bp: Optional[IRobotBreakpoint] = lines.get(lineno)
            if bp:
                # Mark it to stop and then go over exclusions based on condition
                # and hit_condition.
                stop_reason = ReasonEnum.REASON_BREAKPOINT

                if bp.condition:
                    try:
                        from robot.variables.evaluation import (
                            evaluate_expression,
                        )  # noqa

                        curr_vars = ctx.variables.current
                        hit = bool(
                            evaluate_expression(
                                curr_vars.replace_string(bp.condition), curr_vars.store
                            )
                        )
                        if not hit:
                            log.debug(
                                "Breakpoint at %s (%s) skipped (%s evaluated to False)",
                                source,
                                lineno,
                                bp.condition,
                            )
                            stop_reason = None
                    except:
                        log.exception("Error evaluating: %s", bp.condition)

                if stop_reason is not None and bp.hit_condition:
                    bp.hits += 1
                    if bp.hits != bp.hit_condition:
                        log.debug(
                            "Breakpoint at %s (%s) skipped (hit condition: %s evaluated to False)",
                            source,
                            lineno,
                            bp.hit_condition,
                        )
                        stop_reason = None

                if stop_reason is not None:
                    if bp.log_message:
                        curr_vars = ctx.variables.current
                        try:
                            message = curr_vars.replace_string(bp.log_message)
                        except Exception as e:
                            message = (
                                f"Error evaluating: {bp.log_message}.\nError: {e}\n"
                            )

                        if not message.endswith(("\n", "\r")):
                            message += "\n"

                        self.write_message(
                            OutputEvent(
                                body=OutputEventBody(
                                    source=Source(path=source),
                                    line=lineno,
                                    output=message,
                                    category="console",
                                )
                            )
                        )
                        log.debug(
                            "Breakpoint at %s (%s) skipped (due to being a log message breakpoint).",
                            source,
                            lineno,
                        )
                        stop_reason = None

        if stop_reason is None and step_cmd is not None:
            if step_cmd == StepEnum.STEP_IN:
                stop_reason = ReasonEnum.REASON_STEP

            elif step_cmd in (StepEnum.STEP_NEXT, StepEnum.STEP_OUT):
                if len(self._stack_ctx_entries_deque) <= self._stop_on_stack_len:
                    stop_reason = ReasonEnum.REASON_STEP

        if stop_reason is not None:
            self.wait_suspended(stop_reason)

    def _after_run_step(self):
        entry = self._stack_ctx_entries_deque.pop()

        if entry.entry_type == "KEYWORD":
            self._ignore_failures_in_stack.pop()

        if self._is_control_step(entry.entry_type):
            self._stop_on_stack_len -= 1

    def start_suite(self, data, result):
        self._stack_ctx_entries_deque.append(
            _SuiteEntry(data.name, data.source, "SUITE")
        )

    def end_suite(self, data, result):
        self._stack_ctx_entries_deque.pop()

    def start_test(self, data, result):
        self._stack_ctx_entries_deque.append(
            _TestEntry(data.name, data.source, data.lineno, "TEST")
        )

    def end_test(self, data, result):
        self._stack_ctx_entries_deque.pop()

    def log_message(self, message):
        from robotframework_debug_adapter.message_utils import (
            extract_source_and_line_from_message,
        )

        # When debugging show any message in the console (if possible with the
        # current keyword as the source).
        try:
            source = None
            lineno = None

            source_and_line = extract_source_and_line_from_message(message.message)

            if source_and_line is not None:
                source, lineno = source_and_line
                source = Source(path=source)
            else:
                if self._stack_ctx_entries_deque:
                    lineno = 0
                    step_entry: _StepEntry = self._stack_ctx_entries_deque[-1]
                    source = step_entry.source
                    source = Source(path=source)
                    try:
                        lineno = step_entry.lineno
                    except AttributeError:
                        pass

            self.write_message(
                OutputEvent(
                    body=OutputEventBody(
                        source=source,
                        line=lineno,
                        output=f"{message.message}\n",
                        category="console",
                    )
                )
            )
            self._break_on_log_or_system_message(message)
        except:
            log.exception("Error handling log_message.")

    def message(self, message):
        if message.level in ("FAIL", "ERROR"):
            # We also want to show in the console, not just check breaks.
            return self.log_message(message)
        else:
            try:
                self._break_on_log_or_system_message(message)
            except:
                log.exception("Error handling (system) message.")

    def _break_on_log_or_system_message(self, message):
        stop_reason = None
        if message.level == "FAIL":
            if self.break_on_log_failure:
                stop_reason = ReasonEnum.REASON_EXCEPTION
                exc_name = "Suspended due to logged failure: "

        elif message.level == "ERROR":
            if self.break_on_log_error:
                stop_reason = ReasonEnum.REASON_EXCEPTION
                exc_name = "Suspended due to logged error: "

        if stop_reason is not None:
            if self._ignore_failures_in_stack.ignore():
                return

            from robotframework_debug_adapter.message_utils import (
                extract_source_and_line_from_message,
            )

            source_and_line = extract_source_and_line_from_message(message.message)
            entry = None
            if source_and_line is not None:
                source, lineno = source_and_line
                entry = _LogEntry(message.level, source, lineno, "LOG")
                self._stack_ctx_entries_deque.append(entry)

            self._exc_name = exc_name + message.message
            self._exc_description = message.message
            try:
                self.wait_suspended(stop_reason)
            finally:
                self._exc_name = None
                self._exc_description = None

                if entry is not None:
                    self._stack_ctx_entries_deque.pop()


def _patch(
    execution_context_cls, impl, method_name, call_before_method, call_after_method
):

    original_method = getattr(execution_context_cls, method_name)

    @functools.wraps(original_method)
    def new_method(*args, **kwargs):
        call_before_method(*args, **kwargs)
        try:
            ret = original_method(*args, **kwargs)
        finally:
            call_after_method(*args, **kwargs)
        return ret

    setattr(execution_context_cls, method_name, new_method)


class _DebuggerHolder(object):
    _dbg: Optional[IRobotDebugger] = None


def set_global_robot_debugger(dbg: IRobotDebugger):
    _DebuggerHolder._dbg = dbg


def get_global_robot_debugger() -> Optional[IRobotDebugger]:
    return _DebuggerHolder._dbg


def _apply_monkeypatching_latest(impl):
    from robot.running.model import If, For
    from robot.running.bodyrunner import KeywordRunner

    _patch(
        KeywordRunner,
        impl,
        "run",
        impl.before_keyword_runner,
        impl.after_keyword_runner,
    )
    _patch(If, impl, "run", impl.before_control_flow_stmt, impl.after_control_flow_stmt)
    _patch(
        For, impl, "run", impl.before_control_flow_stmt, impl.after_control_flow_stmt
    )


def _apply_monkeypatching_before_4_b_2(impl):
    from robot.running.steprunner import (  # type: ignore
        StepRunner,  #  @UnresolvedImport
    )

    _patch(StepRunner, impl, "run_step", impl.before_run_step, impl.after_run_step)

    try:
        from robot.running.model import For  # type: ignore # @UnresolvedImport

        _patch(
            For,
            impl,
            "run",
            impl.before_control_flow_stmt,
            impl.after_control_flow_stmt,
        )
    except:
        # This may not be the same on older versions...
        pass

    try:
        from robot.running.model import If  # type: ignore # @UnresolvedImport

        _patch(
            If, impl, "run", impl.before_control_flow_stmt, impl.after_control_flow_stmt
        )
    except:
        # This may not be the same on older versions...
        pass


def install_robot_debugger() -> IRobotDebugger:
    """
    Installs the robot debugger and registers it where needed. If a debugger
    is currently installed, resets it (in this case, any existing session,
    stack trace, breakpoints, etc. are reset).
    """

    impl = get_global_robot_debugger()

    if impl is None:
        # Note: only patches once, afterwards, returns the same instance.
        from robotframework_debug_adapter.listeners import DebugListener
        from robotframework_debug_adapter.listeners import DebugListenerV2

        impl = _RobotDebuggerImpl()

        DebugListener.on_start_suite.register(impl.start_suite)
        DebugListener.on_end_suite.register(impl.end_suite)

        DebugListener.on_start_test.register(impl.start_test)
        DebugListener.on_end_test.register(impl.end_test)

        DebugListener.on_log_message.register(impl.log_message)
        DebugListener.on_message.register(impl.message)

        # On RobotFramework 3.x and earlier 4.x dev versions, we do some monkey-patching because
        # the listener was not able to give linenumbers.
        from robot import get_version

        version = get_version()
        use_monkeypatching = version.startswith("3.") or version.startswith("4.0.a")

        if not use_monkeypatching:
            # 4.0.0 onwards
            DebugListenerV2.on_start_keyword.register(impl.start_keyword_v2)
            DebugListenerV2.on_end_keyword.register(impl.end_keyword_v2)
        else:
            # Older versions
            try:
                _apply_monkeypatching_before_4_b_2(impl)
            except ImportError:
                _apply_monkeypatching_latest(impl)

        set_global_robot_debugger(impl)
    else:
        impl.reset()

    return impl



from __future__ import annotations

from functools import partial
import itertools
import threading
from typing import Optional, Dict
import sys

if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol

import typing
import weakref
from robotframework_debug_adapter.constants import DEBUG

from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    TerminatedEvent,
    StoppedEvent,
)


log = get_logger(__name__)

if typing.TYPE_CHECKING:
    from robotframework_debug_adapter.debug_adapter_comm import DebugAdapterComm


class IProtocolMessageCallable(Protocol):
    def __call__(self, message: BaseSchema) -> None:
        pass


class INextSeq(Protocol):
    def __call__(self) -> int:
        pass


class BaseLaunchProcessTargetComm(threading.Thread):
    def __init__(self, debug_adapter_comm: DebugAdapterComm) -> None:
        threading.Thread.__init__(self)
        self._weak_debug_adapter_comm = weakref.ref(debug_adapter_comm)

        self._connected_event = threading.Event()

        self._process_event_msg = None
        self._process_event = threading.Event()

        self._terminated_event_msg: Optional[TerminatedEvent] = None
        self._terminated_lock = threading.Lock()
        self._terminated_event = threading.Event()
        self._next_seq: INextSeq = partial(next, itertools.count(0))
        self._msg_id_to_on_response: Dict[int, Optional[IProtocolMessageCallable]] = {}

    def _handle_received_protocol_message_from_backend(
        self, protocol_message: BaseSchema, backend: str
    ) -> None:
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            READER_THREAD_STOPPED,
        )
        import json

        if protocol_message is READER_THREAD_STOPPED:
            if DEBUG:
                log.debug(
                    f"{self.__class__.__name__} when reading from {backend}: READER_THREAD_STOPPED."
                )
            return

        if DEBUG:
            log.debug(
                "Process json: %s\n"
                % (json.dumps(protocol_message.to_dict(), indent=4, sort_keys=True),)
            )

        try:
            on_response: Optional[IProtocolMessageCallable] = None
            if protocol_message.type == "request":
                from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Request

                req = typing.cast(Request, protocol_message)
                method_name = f"on_{req.command}_request"

            elif protocol_message.type == "event":
                from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Event

                ev = typing.cast(Event, protocol_message)
                method_name = f"on_{ev.event}_event"

            elif protocol_message.type == "response":
                from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Response

                resp = typing.cast(Response, protocol_message)
                on_response = self._msg_id_to_on_response.pop(resp.request_seq, None)
                method_name = f"on_{resp.command}_response"

            else:
                if DEBUG:
                    log.debug(
                        "Unable to decide how to deal with protocol type: %s (read from %s - %s).\n"
                        % (protocol_message.type, backend, self.__class__.__name__)
                    )
                return

            if on_response is not None:
                on_response(protocol_message)

            on_request = getattr(self, method_name, None)

            if on_request is not None:
                on_request(protocol_message)
            elif on_response is not None:
                pass
            else:
                if DEBUG:
                    log.debug(
                        "Unhandled: %s not available when reading from %s - %s.\n"
                        % (method_name, backend, self.__class__.__name__)
                    )
        except:
            log.exception("Error")

    def on_stopped_event(self, event: StoppedEvent):
        debug_adapter_comm = self._weak_debug_adapter_comm()
        if debug_adapter_comm is not None:
            debug_adapter_comm.weak_stopped_target_comm = weakref.ref(self)
            debug_adapter_comm.write_to_client_message(event)
        else:
            log.debug("Command processor collected in event: %s" % (event,))



from robot.api.parsing import ModelTransformer, Token

try:
    from robot.api.parsing import InlineIfHeader
except ImportError:
    InlineIfHeader = None
from robotidy.decorators import check_start_end_line
from robotidy.utils import ROBOT_VERSION


EOL = Token(Token.EOL)
CONTINUATION = Token(Token.CONTINUATION)


class SplitTooLongLine(ModelTransformer):
    """
    Split too long lines.
    If any line in keyword call exceeds given length limit (120 by default) it will be
    split:

        Keyword With Longer Name    ${arg1}    ${arg2}    ${arg3}  # let's assume that arg2 is at 120 char

    To:

        # let's assume that arg2 is at 120 char
        Keyword With Longer Name
        ...    ${arg1}
        ...    ${arg2}
        ...    ${arg3}

    Allowed line length is configurable using global parameter ``--line-length``:

        robotidy --line-length 140 src.robot

    Or using dedicated for this transformer parameter ``line_length``:

        robotidy --configure SplitTooLongLine:line_length:140 src.robot

    Using ``split_on_every_arg`` flag (``True`` by default), you can force the formatter to fill arguments in one line
    until character limit:

        Keyword With Longer Name    ${arg1}
        ...    ${arg2}    ${arg3}

    Supports global formatting params: ``spacecount``, ``separator``, ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/SplitTooLongLine.html for more examples.
    """

    def __init__(self, line_length: int = None, split_on_every_arg: bool = True):
        super().__init__()
        self._line_length = line_length
        self.split_on_every_arg = split_on_every_arg

    @property
    def line_length(self):
        return self.formatting_config.line_length if self._line_length is None else self._line_length

    def visit_If(self, node):  # noqa
        if self.is_inline(node):
            return node
        if node.orelse:
            self.generic_visit(node.orelse)
        return self.generic_visit(node)

    @staticmethod
    def is_inline(node):
        return ROBOT_VERSION.major > 4 and isinstance(node.header, InlineIfHeader)

    @check_start_end_line
    def visit_KeywordCall(self, node):  # noqa
        if all(line[-1].end_col_offset < self.line_length for line in node.lines):
            return node
        return self.split_keyword_call(node)

    @staticmethod
    def join_on_separator(tokens, separator):
        for token in tokens:
            yield token
            yield separator

    @staticmethod
    def split_to_multiple_lines(tokens, indent, separator):
        first = True
        for token in tokens:
            yield indent
            if not first:
                yield CONTINUATION
                yield separator
            yield token
            yield EOL
            first = False

    def split_keyword_call(self, node):
        separator = Token(Token.SEPARATOR, self.formatting_config.separator)
        indent = node.tokens[0]

        split_every_arg = self.split_on_every_arg
        keyword = node.get_token(Token.KEYWORD)
        line = [indent, *self.join_on_separator(node.get_tokens(Token.ASSIGN), separator), keyword]
        if not self.col_fit_in_line(line):
            split_every_arg
            head = [
                *self.split_to_multiple_lines(node.get_tokens(Token.ASSIGN), indent=indent, separator=separator),
                indent,
                CONTINUATION,
                separator,
                keyword,
            ]
            line = []
        else:
            head = []

        comments = []

        # Comments with separators inside them are split into
        # [COMMENT, SEPARATOR, COMMENT] tokens in the AST, so in order to preserve the
        # original comment, we need a lookback on the separator tokens.
        last_separator = None

        rest = node.tokens[node.tokens.index(keyword) + 1 :]
        for token in rest:
            if token.type == Token.SEPARATOR:
                last_separator = token
            elif token.type in {Token.EOL, Token.CONTINUATION}:
                continue
            elif token.type == Token.COMMENT:
                # AST splits comments with separators, e.g.
                #
                # "# Comment     rest" -> ["# Comment", "     ", "rest"].
                #
                # Notice the third value not starting with a hash - that's what this
                # condition is about:
                if not str(token).startswith("#"):
                    # -2 because -1 is the EOL
                    comments[-2].value += last_separator.value + token.value
                else:
                    comments += [indent, token, EOL]
            elif token.type == Token.ARGUMENT:
                if token.value == "":
                    token.value = "${EMPTY}"
                if self.split_on_every_arg or not self.col_fit_in_line(line + [separator, token]):
                    line.append(EOL)
                    head += line
                    line = [indent, CONTINUATION, separator, token]
                else:
                    line += [separator, token]

        # last line
        line.append(EOL)
        head += line

        node.tokens = comments + head
        return node

    def col_fit_in_line(self, tokens):
        return self.len_token_text(tokens) < self.line_length

    @staticmethod
    def len_token_text(tokens):
        return sum(len(token.value) for token in tokens)



try:
    from PySide import QtCore
except:
    try:
        from PySide2 import QtCore
    except:
        try:
            from PyQt4 import QtCore
        except:
            from PyQt5 import QtCore


class TestObject(QtCore.QObject):
    """
    Test class providing some non-argument signal
    """

    try:
        testSignal = QtCore.Signal()  # @UndefinedVariable
    except:
        testSignal = QtCore.pyqtSignal()  # @UndefinedVariable


class TestThread(QtCore.QThread):

    def run(self):
        QtCore.QThread.sleep(4)
        print('Done sleeping')


def on_start():
    print('On start called1')
    print('On start called2')


app = QtCore.QCoreApplication([])
some_thread = TestThread()
some_object = TestObject()

# connect QThread.started to the signal
some_thread.started.connect(some_object.testSignal)
some_object.testSignal.connect(on_start)
some_thread.finished.connect(app.quit)

some_thread.start()
app.exec_()
print('TEST SUCEEDED!')



from robotframework_ls.impl.protocols import (
    ICompletionContext,
    EvaluatableExpressionTypedDict,
    IRobotToken,
)
from typing import Optional


def _create_evaluatable_expression_from_token(
    token: IRobotToken, expression: str
) -> EvaluatableExpressionTypedDict:
    return {
        "range": {
            "start": {"line": token.lineno - 1, "character": token.col_offset},
            "end": {
                "line": token.lineno - 1,
                "character": token.end_col_offset,
            },
        },
        "expression": expression,
    }


def provide_evaluatable_expression(
    completion_context: ICompletionContext,
) -> Optional[EvaluatableExpressionTypedDict]:
    from robotframework_ls.impl import ast_utils

    var_token_info = completion_context.get_current_variable()
    if var_token_info is not None:
        token = var_token_info.token
        return _create_evaluatable_expression_from_token(
            token, var_token_info.var_identifier + "{" + token.value + "}"
        )

    token_info = completion_context.get_current_token()
    if token_info is not None:
        keyword_name_token = ast_utils.get_keyword_name_token(
            token_info.stack, token_info.node, token_info.token
        )
        if keyword_name_token is not None:
            return _create_evaluatable_expression_from_token(
                keyword_name_token, keyword_name_token.value
            )

    return None



from typing import Iterable

from robot.api.parsing import ModelTransformer, Token

try:
    from robot.api.parsing import Continue, Break
except ImportError:
    Continue, Break = None, None

from robotidy.utils import normalize_name, after_last_dot, wrap_in_if_and_replace_statement, ROBOT_VERSION
from robotidy.decorators import check_start_end_line


class ReplaceBreakContinue(ModelTransformer):
    """
    Replace Continue For Loop and Exit For Loop keyword variants with CONTINUE and BREAK statements.

    Following code:

        *** Keywords ***
        Keyword
            FOR    ${var}    IN  1  2
                Continue For Loop
                Continue For Loop If    $condition
                Exit For Loop
                Exit For Loop If    $condition
            END

    will be transformed to:

        *** Keywords ***
        Keyword
            FOR    ${var}    IN  1  2
                CONTINUE
                IF    $condition
                    CONTINUE
                END
                BREAK
                IF    $condition
                    BREAK
                END
            END

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/ReplaceBreakContinue.html for more examples.
    """

    ENABLED = ROBOT_VERSION.major >= 5

    def __init__(self):
        self.in_loop = False

    def visit_File(self, node):  # noqa
        self.in_loop = False
        return self.generic_visit(node)

    @staticmethod
    def create_statement_from_tokens(statement, tokens: Iterable, indent: Token):
        return statement([indent, Token(statement.type), *tokens])

    @check_start_end_line
    def visit_KeywordCall(self, node):  # noqa
        if not self.in_loop or not node.keyword or node.errors:
            return node
        normalized_name = after_last_dot(normalize_name(node.keyword))
        if "forloop" not in normalized_name:
            return node
        if normalized_name == "continueforloop":
            return self.create_statement_from_tokens(statement=Continue, tokens=node.tokens[2:], indent=node.tokens[0])
        elif normalized_name == "exitforloop":
            return self.create_statement_from_tokens(statement=Break, tokens=node.tokens[2:], indent=node.tokens[0])
        elif normalized_name == "continueforloopif":
            return wrap_in_if_and_replace_statement(node, Continue, self.formatting_config.separator)
        elif normalized_name == "exitforloopif":
            return wrap_in_if_and_replace_statement(node, Break, self.formatting_config.separator)
        return node

    def visit_For(self, node):  # noqa
        self.in_loop = True
        node = self.generic_visit(node)
        self.in_loop = False
        return node

    visit_While = visit_For



from typing import Any, Dict, List, Optional, Deque
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    StartSuiteEvent,
    StartSuiteEventBody,
    StartTestEvent,
    StartTestEventBody,
    EndSuiteEvent,
    EndSuiteEventBody,
    EndTestEvent,
    EndTestEventBody,
)
from collections import namedtuple
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


def send_event(event):
    from robotframework_debug_adapter.global_vars import get_global_robot_target_comm

    robot_target_comm = get_global_robot_target_comm()
    if robot_target_comm is not None:
        robot_target_comm.write_message(event)


_SourceInfo = namedtuple("_SourceInfo", "source, lineno, test_name")


class _EventsState:
    def __init__(self):
        from collections import deque
        from robotframework_debug_adapter._ignore_failures_in_stack import (
            IgnoreFailuresInStack,
        )

        self._failed_keywords: Optional[List[Dict[str, Any]]] = None
        self._failure_messages: List[str] = []
        self._add_to_test_failure: str = ""
        self._source_info_stack: Deque[_SourceInfo] = deque()
        self._ignore_failures_in_stack = IgnoreFailuresInStack()


_global_events_state = _EventsState()


def _get_events_state():
    return _global_events_state


class EventsListenerV3:
    # Note: see https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html
    # for actual attributes.
    ROBOT_LISTENER_API_VERSION = 3

    def __init__(self):
        # Whenever it's created, reset the global state.
        global _global_events_state
        _global_events_state = _EventsState()

    # start suite/test
    def start_suite(self, data, result) -> None:
        source = data.source
        name = data.name

        tests = []
        for test in data.tests:
            tests.append(test.name)

        state = _get_events_state()
        state._source_info_stack.append(_SourceInfo(source, None, None))

        send_event(StartSuiteEvent(StartSuiteEventBody(name, source, tests)))

    def start_test(self, data, result) -> None:
        state = _get_events_state()

        if state._failure_messages:
            state._add_to_test_failure = "\n".join(state._failure_messages)
            state._failure_messages = []

        name = data.name
        source = data.source
        lineno = data.lineno
        state._source_info_stack.append(_SourceInfo(source, lineno, name))

        send_event(StartTestEvent(StartTestEventBody(name, source, lineno)))

    # end suite/test

    def end_suite(self, data, result) -> None:
        state = _get_events_state()
        try:
            send_event(
                EndSuiteEvent(
                    EndSuiteEventBody(
                        data.name,
                        elapsedtime=result.elapsedtime,
                        status=result.status,
                        source=data.source,
                        message=result.message.strip(),
                        failed_keywords=state._failed_keywords,
                    )
                )
            )
        finally:
            state._failed_keywords = None
            state._source_info_stack.pop()

    def end_test(self, data, result) -> None:
        state = _get_events_state()
        try:
            msg = result.message.strip()
            if state._add_to_test_failure:
                msg = state._add_to_test_failure + "\n" + msg
                state._add_to_test_failure = ""

            if state._failure_messages:
                if msg:
                    msg += "\n"
                msg += "\n".join(state._failure_messages)
                state._failure_messages = []

            send_event(
                EndTestEvent(
                    EndTestEventBody(
                        data.name,
                        elapsedtime=result.elapsedtime,
                        status=result.status,
                        source=data.source,
                        message=msg.strip(),
                        failed_keywords=state._failed_keywords,
                    )
                )
            )
        finally:
            state._failed_keywords = None
            state._source_info_stack.pop()


class EventsListenerV2:
    # Note: see https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html
    # for actual attributes.
    ROBOT_LISTENER_API_VERSION = 2

    # For keywords we're just interested on tracking failures to send when a test/suite finished.
    # Note: we also try to capture the failure through logged messages.

    def log_message(self, message: Dict[str, Any]) -> None:
        """
        Called when an executed keyword writes a log message.

        message is a dictionary with the following contents:

            message: The content of the message.
            level: Log level used in logging the message.
            timestamp: Message creation time in format YYYY-MM-DD hh:mm:ss.mil.
            html: String yes or no denoting whether the message should be interpreted as HTML or not.

        Not called if the message level is below the current threshold level.
        """

        message_string = message.get("message")
        if not message_string:
            return

        state = _get_events_state()

        if state._ignore_failures_in_stack.ignore():
            return

        from robotframework_debug_adapter.message_utils import (
            extract_source_and_line_from_message,
        )

        level = message.get("level")

        source = None
        lineno = None
        test_name = None

        source_and_line = extract_source_and_line_from_message(message_string)
        if source_and_line is not None:
            source, lineno = source_and_line

        if state._source_info_stack:
            from robotframework_debug_adapter import file_utils

            source_info: _SourceInfo = state._source_info_stack[-1]
            test_name = source_info.test_name  # May be None.

            if source is None:
                source = source_info.source
                source = file_utils.get_abs_path_real_path_and_base_from_file(source)[0]

            if lineno is None:
                lineno = 0
                try:
                    lineno = source_info.lineno
                except AttributeError:
                    pass

        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            LogMessageEvent,
            LogMessageEventBody,
        )

        send_event(
            LogMessageEvent(
                body=LogMessageEventBody(
                    source=source,
                    lineno=lineno,
                    message=f"{message_string}",
                    level=level,
                    testName=test_name,
                )
            )
        )

        if message["level"] in ("FAIL", "ERROR"):  # FAIL/WARN/INFO/DEBUG/TRACE
            state._failure_messages.append(message["message"])

    def message(self, message):
        if message["level"] in ("FAIL", "ERROR"):
            # We also want to show these for system messages.
            return self.log_message(message)

    def start_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        state = _get_events_state()
        state._ignore_failures_in_stack.push(attributes.get("kwname", ""))
        source = attributes.get("source")
        lineno = attributes.get("lineno")
        test_name = None
        if state._source_info_stack:
            # Keep same test name
            test_name = state._source_info_stack[-1].test_name
        state._source_info_stack.append(_SourceInfo(source, lineno, test_name))

    def end_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
        try:
            state = _get_events_state()
            status = attributes.get("status")
            # Status could be PASS, FAIL, SKIP or NOT RUN. SKIP and NOT RUN
            if status == "FAIL":

                if state._ignore_failures_in_stack.ignore():
                    return

                if state._failed_keywords is None:
                    state._failed_keywords = []
                source = attributes.get("source")
                if not source:
                    return
                lineno = attributes.get("lineno")
                if lineno is None:
                    return

                state._failed_keywords.append(
                    {
                        "name": name,
                        "source": source,
                        "lineno": lineno,
                        "failure_messages": state._failure_messages,
                    }
                )
                state._failure_messages = []
        finally:
            try:
                state._source_info_stack.pop()
            except:
                log.exception("Error in state._source_info_stack.pop()")

            state._ignore_failures_in_stack.pop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# path.py: Path functions.
#
# Copyright (C) 2010 Yesudeep Mangalapilly <yesudeep@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
:module: pathtools.path
:synopsis: Directory walking, listing, and path sanitizing functions.
:author: Yesudeep Mangalapilly <yesudeep@gmail.com>

Functions
---------
.. autofunction:: get_dir_walker
.. autofunction:: walk
.. autofunction:: listdir
.. autofunction:: list_directories
.. autofunction:: list_files
.. autofunction:: absolute_path
.. autofunction:: real_absolute_path
.. autofunction:: parent_dir_path
"""

import os.path
import os.path
from functools import partial


__all__ = [
    'get_dir_walker',
    'walk',
    'listdir',
    'list_directories',
    'list_files',
    'absolute_path',
    'real_absolute_path',
    'parent_dir_path',
]


def get_dir_walker(recursive, topdown=True, followlinks=False):
    """
    Returns a recursive or a non-recursive directory walker.

    :param recursive:
        ``True`` produces a recursive walker; ``False`` produces a non-recursive
        walker.
    :returns:
        A walker function.
    """
    if recursive:
        walk = partial(os.walk, topdown=topdown, followlinks=followlinks)
    else:
        def walk(path, topdown=topdown, followlinks=followlinks):
            try:
                yield next(os.walk(path, topdown=topdown, followlinks=followlinks))
            except NameError:
                yield os.walk(path, topdown=topdown, followlinks=followlinks).next() #IGNORE:E1101
    return walk


def walk(dir_pathname, recursive=True, topdown=True, followlinks=False):
    """
    Walks a directory tree optionally recursively. Works exactly like
    :func:`os.walk` only adding the `recursive` argument.

    :param dir_pathname:
        The directory to traverse.
    :param recursive:
        ``True`` for walking recursively through the directory tree;
        ``False`` otherwise.
    :param topdown:
        Please see the documentation for :func:`os.walk`
    :param followlinks:
        Please see the documentation for :func:`os.walk`
    """
    walk_func = get_dir_walker(recursive, topdown, followlinks)
    for root, dirnames, filenames in walk_func(dir_pathname):
        yield (root, dirnames, filenames)


def listdir(dir_pathname,
            recursive=True,
            topdown=True,
            followlinks=False):
    """
    Enlists all items using their absolute paths in a directory, optionally
    recursively.

    :param dir_pathname:
        The directory to traverse.
    :param recursive:
        ``True`` for walking recursively through the directory tree;
        ``False`` otherwise.
    :param topdown:
        Please see the documentation for :func:`os.walk`
    :param followlinks:
        Please see the documentation for :func:`os.walk`
    """
    for root, dirnames, filenames\
    in walk(dir_pathname, recursive, topdown, followlinks):
        for dirname in dirnames:
            yield absolute_path(os.path.join(root, dirname))
        for filename in filenames:
            yield absolute_path(os.path.join(root, filename))


def list_directories(dir_pathname,
                     recursive=True,
                     topdown=True,
                     followlinks=False):
    """
    Enlists all the directories using their absolute paths within the specified
    directory, optionally recursively.

    :param dir_pathname:
        The directory to traverse.
    :param recursive:
        ``True`` for walking recursively through the directory tree;
        ``False`` otherwise.
    :param topdown:
        Please see the documentation for :func:`os.walk`
    :param followlinks:
        Please see the documentation for :func:`os.walk`
    """
    for root, dirnames, filenames\
    in walk(dir_pathname, recursive, topdown, followlinks):
        for dirname in dirnames:
            yield absolute_path(os.path.join(root, dirname))


def list_files(dir_pathname,
               recursive=True,
               topdown=True,
               followlinks=False):
    """
    Enlists all the files using their absolute paths within the specified
    directory, optionally recursively.

    :param dir_pathname:
        The directory to traverse.
    :param recursive:
        ``True`` for walking recursively through the directory tree;
        ``False`` otherwise.
    :param topdown:
        Please see the documentation for :func:`os.walk`
    :param followlinks:
        Please see the documentation for :func:`os.walk`
    """
    for root, dirnames, filenames\
    in walk(dir_pathname, recursive, topdown, followlinks):
        for filename in filenames:
            yield absolute_path(os.path.join(root, filename))


def absolute_path(path):
    """
    Returns the absolute path for the given path and normalizes the path.

    :param path:
        Path for which the absolute normalized path will be found.
    :returns:
        Absolute normalized path.
    """
    return os.path.abspath(os.path.normpath(path))


def real_absolute_path(path):
    """
    Returns the real absolute normalized path for the given path.

    :param path:
        Path for which the real absolute normalized path will be found.
    :returns:
        Real absolute normalized path.
    """
    return os.path.realpath(absolute_path(path))


def parent_dir_path(path):
    """
    Returns the parent directory path.

    :param path:
        Path for which the parent directory will be obtained.
    :returns:
        Parent directory path.
    """
    return absolute_path(os.path.dirname(path))



import argparse
from typing import Pattern, Set, Dict
import fnmatch
import os
import re
import sys
from itertools import chain
from pathlib import Path

import toml
from robot.utils import FileReader

from robocop.exceptions import (
    ArgumentFileNotFoundError,
    ConfigGeneralError,
    InvalidArgumentError,
    NestedArgumentFileError,
)
from robocop.rules import RuleSeverity
from robocop.files import DEFAULT_EXCLUDES, find_file_in_project_root, find_project_root
from robocop.utils import RecommendationFinder
from robocop.version import __version__


def translate_pattern(pattern: str) -> Pattern:
    return re.compile(fnmatch.translate(pattern))


class ParseDelimitedArgAction(argparse.Action):  # pylint: disable=too-few-public-methods
    def __call__(self, parser, namespace, values, option_string=None):
        container = getattr(namespace, self.dest)
        container.update(values.split(","))


class ParseCheckerConfig(argparse.Action):  # pylint: disable=too-few-public-methods
    def __call__(self, parser, namespace, values, option_string=None):
        container = getattr(namespace, self.dest)
        container.append(values.strip())


class ParseFileTypes(argparse.Action):  # pylint: disable=too-few-public-methods
    def __call__(self, parser, namespace, values, option_string=None):
        filetypes = getattr(namespace, self.dest)
        for filetype in values.split(","):
            filetypes.add(filetype if filetype.startswith(".") else "." + filetype)
        setattr(namespace, self.dest, filetypes)


class SetRuleThreshold(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(
            namespace, self.dest, RuleSeverity.parser(values) if values in ("I", "W", "E") else RuleSeverity.INFO
        )  # TODO Change severity to enum


class SetListOption(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        pattern = values if values else "*"
        if "*" in pattern:
            pattern = translate_pattern(pattern)
        setattr(namespace, self.dest, pattern)


class CustomArgParser(argparse.ArgumentParser):
    def __init__(self, *args, from_cli=False, **kwargs):
        self.from_cli = from_cli
        super().__init__(*args, **kwargs)

    def error(self, message):
        if self.from_cli:
            super().error(message)
        else:
            raise InvalidArgumentError(message)


def validate_regex(pattern: str) -> Pattern:
    try:
        return re.compile(pattern) if pattern is not None else None
    except re.error:
        raise ConfigGeneralError(f"Provided regex pattern {pattern} failed to be compiled")


class Config:
    def __init__(self, root=None, from_cli: bool = False):
        self.from_cli = from_cli
        self.exec_dir = os.path.abspath(".")
        self.root = Path(root) if root is not None else root
        self.include = set()
        self.exclude = set()
        self.ignore = set()
        self.ignore_default = re.compile(DEFAULT_EXCLUDES)
        self.reports = {"return_status"}
        self.threshold = RuleSeverity("I")
        self.configure = []
        self.format = "{source}:{line}:{col} [{severity}] {rule_id} {desc} ({name})"
        self.paths = ["."]
        self.ext_rules = set()
        self.include_patterns = []
        self.exclude_patterns = []
        self.filetypes = {".robot", ".resource", ".tsv"}
        self.list = ""
        self.list_configurables = ""
        self.list_reports = False
        self.output = None
        self.recursive = True
        self.verbose = False
        self.config_from = ""
        self.parser = self._create_parser()
        self.parse_opts(from_cli=from_cli)

    def remove_severity(self):
        self.include = {self.replace_severity_values(rule) for rule in self.include}
        self.exclude = {self.replace_severity_values(rule) for rule in self.exclude}
        for index, conf in enumerate(self.configure):
            if conf.count(":") != 2:
                continue
            message, param, value = conf.split(":")
            message = self.replace_severity_values(message)
            self.configure[index] = f"{message}:{param}:{value}"

    @staticmethod
    def filter_patterns_from_names(only_names, only_patterns):
        filtered = set()
        for rule in only_names:
            if "*" in rule:
                only_patterns.append(translate_pattern(rule))
            else:
                filtered.add(rule)
        return filtered

    def translate_patterns(self):
        self.include = self.filter_patterns_from_names(self.include, self.include_patterns)
        self.exclude = self.filter_patterns_from_names(self.exclude, self.exclude_patterns)

    def preparse(self, args):
        args = sys.argv[1:] if args is None else args
        parsed_args = []
        args = (arg for arg in args)
        for arg in args:
            if arg in ("-A", "--argumentfile"):
                try:
                    argfile = next(args)
                except StopIteration:
                    raise ArgumentFileNotFoundError("") from None
                parsed_args += self.load_args_from_file(argfile)
            else:
                parsed_args.append(arg)
        return parsed_args

    def load_args_from_file(self, argfile):
        try:
            with FileReader(argfile) as arg_f:
                args = []
                for line in arg_f.readlines():
                    if line.strip().startswith("#"):
                        continue
                    for arg in line.split(" ", 1):
                        arg = arg.strip()
                        if not arg:
                            continue
                        args.append(arg)
                if "-A" in args or "--argumentfile" in args:
                    raise NestedArgumentFileError(argfile)
                if args:
                    self.config_from = argfile
                return args
        except FileNotFoundError:
            raise ArgumentFileNotFoundError(argfile) from None

    def _create_parser(self):
        parser = CustomArgParser(
            prog="robocop",
            formatter_class=argparse.RawTextHelpFormatter,
            description="Static code analysis tool for Robot Framework",
            epilog="For full documentation visit: https://robocop.readthedocs.io/en/latest/",
            add_help=False,
            from_cli=self.from_cli,
        )
        required = parser.add_argument_group(title="Required parameters")
        optional = parser.add_argument_group(title="Optional parameters")

        required.add_argument(
            "paths",
            metavar="paths",
            type=str,
            nargs="*",
            default=self.paths,
            help="List of paths (files or directories) to be parsed by Robocop.",
        )
        optional.add_argument(
            "-i",
            "--include",
            action=ParseDelimitedArgAction,
            default=self.include,
            metavar="RULES",
            help="Run Robocop only with specified rules. You can define rule by its name or id.\n"
            "Glob patterns are supported.",
        )
        optional.add_argument(
            "-e",
            "--exclude",
            action=ParseDelimitedArgAction,
            default=self.exclude,
            metavar="RULES",
            help="Ignore specified rules. You can define rule by its name or id.\nGlob patterns are supported.",
        )
        optional.add_argument(
            "-rules",
            "--ext-rules",
            action=ParseDelimitedArgAction,
            default=self.ext_rules,
            help="List of paths with custom rules.",
        )
        optional.add_argument(
            "-nr",
            "--no-recursive",
            dest="recursive",
            action="store_false",
            help="Use this flag to stop scanning directories recursively.",
        )
        optional.add_argument(
            "-r",
            "--reports",
            action=ParseDelimitedArgAction,
            default=self.reports,
            help="Generate reports after scan.\n"
            "You can enable reports by listing them in comma-separated list:\n"
            "--reports rules_by_id,rules_by_error_type,scan_timer\n"
            "To enable all reports use all:\n"
            "--report all",
        )
        optional.add_argument(
            "-f",
            "--format",
            type=str,
            default=self.format,
            help="Format of output message. "
            "You can use placeholders to change the way an issue is reported.\n"
            "Default: {source}:{line}:{col} [{severity}] {rule_id} {desc} ({name})",
        )
        optional.add_argument(
            "-c",
            "--configure",
            action=ParseCheckerConfig,
            default=self.configure,
            metavar="CONFIGURABLE",
            help="Configure checker or report with parameter value. Usage:\n"
            "-c message_name_or_id:param_name:param_value\n"
            "Examples:\n"
            "-c line-too-long:line_length:150\n"
            "--configure 0101:severity:E",
        )
        optional.add_argument(
            "-l",
            "--list",
            action=SetListOption,
            nargs="?",
            const="",
            default=self.list,
            metavar="PATTERN",
            help="List all available rules. You can use optional PATTERN argument.",
        )
        optional.add_argument(
            "-lc",
            "--list-configurables",
            action=SetListOption,
            nargs="?",
            const="",
            default=self.list_configurables,
            metavar="PATTERN",
            help="List all available rules with configurable parameters. " "You can use optional PATTERN argument.",
        )
        optional.add_argument(
            "-lr",
            "--list-reports",
            action="store_true",
            default=self.list_reports,
            help="List all available reports.",
        )
        optional.add_argument(
            "-o",
            "--output",
            type=argparse.FileType("w"),
            default=self.output,
            metavar="PATH",
            help="Path to output file.",
        )
        optional.add_argument(
            "-ft",
            "--filetypes",
            action=ParseFileTypes,
            default=self.filetypes,
            help="Comma-separated list of file extensions to be scanned by Robocop",
        )
        optional.add_argument(
            "-t",
            "--threshold",
            action=SetRuleThreshold,
            default=self.threshold,
            help=f"Disable rules below given threshold. Available message levels: "
            f'{" < ".join(sev.value for sev in RuleSeverity)}',
        )
        optional.add_argument("-A", "--argumentfile", metavar="PATH", help="Path to file with arguments.")
        optional.add_argument(
            "-g",
            "--ignore",
            action=ParseDelimitedArgAction,
            default=self.ignore,
            metavar="PATH",
            help="Ignore file(s) and path(s) provided. Glob patterns are supported.",
        )
        optional.add_argument(
            "-gd",
            "--ignore-default",
            type=validate_regex,
            default=self.ignore_default,
            metavar="PATTERN",
            help=f"Paths ignored by default. "
            f"A regular expression to exclude directories on file search.\n"
            f"An empty value means no path is excluded. Default: {DEFAULT_EXCLUDES}",
        )
        optional.add_argument("-h", "--help", action="help", help="Print this help message and exit.")
        optional.add_argument(
            "-v",
            "--version",
            action="version",
            version=__version__,
            help="Display Robocop version.",
        )
        optional.add_argument(
            "-vv", "--verbose", action="store_true", help="Display extra information during execution."
        )
        optional.add_argument(
            "--directives",
            action="version",
            version="1. Serve the public trust\n2. Protect the innocent\n3. Uphold the law\n4. [ACCESS " "DENIED]",
            help=argparse.SUPPRESS,
        )

        return parser

    def parse_opts(self, args=None, from_cli: bool = True):
        if self.root is None:
            self.root = find_project_root(self.paths)
        default_args = self.load_default_config_file()
        if default_args is None:
            self.load_pyproject_file()
        else:
            default_args = self.preparse(default_args)
            self.parse_args_to_config(default_args)

        args = self.preparse(args) if from_cli else None
        if args:
            args = self.parse_args_to_config(args)

        self.remove_severity()
        self.translate_patterns()

        if self.verbose:
            if self.config_from:
                print(f"Loaded configuration from {self.config_from}")
            else:
                print("No config file found or configuration is empty. Using default configuration")

        return args

    def load_default_config_file(self):
        robocop_path = find_file_in_project_root(".robocop", self.root)
        if robocop_path.is_file():
            return self.load_args_from_file(robocop_path)
        return None

    def load_pyproject_file(self):
        pyproject_path = find_file_in_project_root("pyproject.toml", self.root)
        if not pyproject_path.is_file():
            return
        try:
            config = toml.load(str(pyproject_path))
        except toml.TomlDecodeError as err:
            raise InvalidArgumentError(f"Failed to decode {str(pyproject_path)}: {err}") from None
        config = config.get("tool", {}).get("robocop", {})
        if self.parse_toml_to_config(config):
            self.config_from = pyproject_path

    @staticmethod
    def replace_in_set(container: Set, old_key: str, new_key: str):
        if old_key not in container:
            return
        container.remove(old_key)
        container.add(new_key)

    def validate_rule_names(self, rules):
        # add rule name in form of old_name: new_name
        deprecated = {}
        for rule in chain(self.include, self.exclude):
            if rule in deprecated:  # update warning description to specific case
                print(
                    f"### DEPRECATION WARNING ###\nThe name (or ID) of the rule '{rule}' is "
                    f"renamed to '{deprecated[rule]}'. "
                    f"Update your configuration if you're using old name. "
                    f"This information will disappear in the next version (X.Y.Z)\n\n"
                )
                self.replace_in_set(self.include, rule, deprecated[rule])
                self.replace_in_set(self.exclude, rule, deprecated[rule])
            elif rule not in rules:
                similar = RecommendationFinder().find_similar(rule, rules)
                raise ConfigGeneralError(f"Provided rule '{rule}' does not exist. {similar}")

    def is_rule_enabled(self, rule):
        if self.is_rule_disabled(rule):
            return False
        if self.include or self.include_patterns:  # if any include pattern, it must match with something
            if rule.rule_id in self.include or rule.name in self.include:
                return True
            for pattern in self.include_patterns:
                if pattern.match(rule.rule_id) or pattern.match(rule.name):
                    return True
            return False
        return True

    def is_rule_disabled(self, rule):
        if not rule.enabled_in_version:
            return True
        if rule.severity < self.threshold:
            return True
        if rule.rule_id in self.exclude or rule.name in self.exclude:
            return True
        for pattern in self.exclude_patterns:
            if pattern.match(rule.rule_id) or pattern.match(rule.name):
                return True
        return False

    def is_path_ignored(self, path):
        for pattern in self.ignore:
            if path.match(pattern):
                return True
        if self.ignore_default:
            match = self.ignore_default.search(str(path))
            return bool(match and match.group(0))
        return False

    @staticmethod
    def replace_severity_values(rule_name: str):
        sev = "".join(sev.value for sev in RuleSeverity)
        if re.match(f"[{sev}][0-9]{{4,}}", rule_name):
            for char in sev:
                rule_name = rule_name.replace(char, "")
        return rule_name

    def parse_toml_to_config(self, toml_data: Dict):
        if not toml_data:
            return False
        assign_type = {"paths", "format"}
        set_type = {"include", "exclude", "reports", "ignore", "ext_rules"}
        append_type = {"configure"}
        toml_data = {key.replace("-", "_"): value for key, value in toml_data.items()}
        for key, value in toml_data.items():
            if key in assign_type:
                self.__dict__[key] = value
            elif key in set_type:
                self.__dict__[key].update(set(value))
            elif key in append_type:
                self.__dict__[key] += value
            elif key == "filetypes":
                for filetype in toml_data["filetypes"]:
                    self.filetypes.add(filetype if filetype.startswith(".") else "." + filetype)
            elif key == "threshold":
                self.threshold = RuleSeverity(value)
            elif key == "output":
                self.output = open(value, "w")
            elif key == "no_recursive":
                self.recursive = not value
            elif key == "verbose":
                self.verbose = value
            else:
                raise InvalidArgumentError(f"Option '{key}' is not supported in pyproject.toml configuration file.")
        return True

    def parse_args_to_config(self, args):
        if args is None:
            return None

        args = self.parser.parse_args(args)
        for key, value in dict(**vars(args)).items():
            if key in self.__dict__:
                self.__dict__[key] = value

        return args



import time
import subprocess
import sys
import os

if __name__ == '__main__':
    if 'launch-subprocesses' in sys.argv:
        n = int(sys.argv[-1])
        if n != 0:
            subprocess.Popen([sys.executable, __file__, 'launch-subprocesses', str(n - 1)])
        if hasattr(os, 'getppid'):
            print('%screated %s (child of %s)' % ('\t' * (4 - n), os.getpid(), os.getppid()))
        else:
            print('%screated %s' % ('\t' * (4 - n), os.getpid()))

    elif 'check-subprocesses' in sys.argv or 'check-subprocesses-ignore-pid' in sys.argv:
        # Recursively create a process tree such as:
        # - parent (this process)
        #    - p3
        #      - p2
        #        - p1
        #          - p0
        #    - p3
        #      - p2
        #        - p1
        #          - p0
        p0 = subprocess.Popen([sys.executable, __file__, 'launch-subprocesses', '3'])
        p1 = subprocess.Popen([sys.executable, __file__, 'launch-subprocesses', '3'])

        if 'check-subprocesses-ignore-pid' in sys.argv:
            import pydevd
            pydevd.add_dont_terminate_child_pid(p0.pid)

        print('created', os.getpid())

    while True:
        time.sleep(.1)



import sys
import os
from _pydev_bundle import pydev_monkey
sys.path.insert(0, os.path.split(os.path.split(__file__)[0])[0])

import unittest

try:
    import thread
except:
    import _thread as thread  # @UnresolvedImport


#=======================================================================================================================
# TestCase
#=======================================================================================================================
class TestCase(unittest.TestCase):

    def test_start_new_thread(self):
        pydev_monkey.patch_thread_modules()
        try:
            found = {}

            def function(a, b, *args, **kwargs):
                found['a'] = a
                found['b'] = b
                found['args'] = args
                found['kwargs'] = kwargs

            thread.start_new_thread(function, (1, 2, 3, 4), {'d':1, 'e':2})
            import time
            for _i in range(20):
                if len(found) == 4:
                    break
                time.sleep(.1)
            else:
                raise AssertionError('Could not get to condition before 2 seconds')

            self.assertEqual({'a': 1, 'b': 2, 'args': (3, 4), 'kwargs': {'e': 2, 'd': 1}}, found)
        finally:
            pydev_monkey.undo_patch_thread_modules()

    def test_start_new_thread2(self):
        pydev_monkey.patch_thread_modules()
        try:
            found = {}

            class F(object):
                start_new_thread = thread.start_new_thread

                def start_it(self):
                    try:
                        self.start_new_thread(self.function, (1, 2, 3, 4), {'d':1, 'e':2})
                    except:
                        import traceback;traceback.print_exc()

                def function(self, a, b, *args, **kwargs):
                    found['a'] = a
                    found['b'] = b
                    found['args'] = args
                    found['kwargs'] = kwargs

            f = F()
            f.start_it()
            import time
            for _i in range(20):
                if len(found) == 4:
                    break
                time.sleep(.1)
            else:
                raise AssertionError('Could not get to condition before 2 seconds')

            self.assertEqual({'a': 1, 'b': 2, 'args': (3, 4), 'kwargs': {'e': 2, 'd': 1}}, found)
        finally:
            pydev_monkey.undo_patch_thread_modules()


#=======================================================================================================================
# main
#=======================================================================================================================
if __name__ == '__main__':
    unittest.main()



from collections import defaultdict

from robot.api.parsing import ModelTransformer, Token
from robot.parsing.model import Statement

from robotidy.utils import node_outside_selection, round_to_four, tokens_by_lines, left_align, is_blank_multiline


class AlignSettingsSection(ModelTransformer):
    """
    Align statements in ``*** Settings ***`` section to columns.

    Following code::

        *** Settings ***
        Library      SeleniumLibrary
        Library   Mylibrary.py
        Variables  variables.py
        Test Timeout  1 min
            # this should be left aligned

    will be transformed to::

        *** Settings ***
        Library         SeleniumLibrary
        Library         Mylibrary.py
        Variables       variables.py
        Test Timeout    1 min
        # this should be left aligned

    You can configure how many columns should be aligned to longest token in given column. The remaining columns
    will use fixed length separator length ``--spacecount``. By default only first two columns are aligned.
    To align first three columns::

       robotidy --transform AlignSettingsSection:up_to_column=3

    To align all columns set ``up_to_column`` to 0.

    Arguments inside keywords in Suite Setup, Suite Teardown, Test Setup and Test Teardown are indented by additional
    ``argument_indent`` (default ``4``) spaces::

        *** Settings ***
        Suite Setup         Start Session
        ...                     host=${IPADDRESS}
        ...                     user=${USERNAME}
        ...                     password=${PASSWORD}
        Suite Teardown      Close Session

    To disable it configure ``argument_indent`` with ``0``.

    Supports global formatting params: ``--startline``, ``--endline`` and ``--spacecount``
    (for columns with fixed length).

    See https://robotidy.readthedocs.io/en/latest/transformers/AlignSettingsSection.html for more examples.
    """

    TOKENS_WITH_KEYWORDS = {
        Token.SUITE_SETUP,
        Token.SUITE_TEARDOWN,
        Token.TEST_SETUP,
        Token.TEST_TEARDOWN,
    }

    def __init__(self, up_to_column: int = 2, argument_indent: int = 4, min_width: int = None):
        self.up_to_column = up_to_column - 1
        self.argument_indent = argument_indent
        self.min_width = min_width

    def visit_SettingSection(self, node):  # noqa
        if node_outside_selection(node, self.formatting_config):
            return node
        statements = []
        for child in node.body:
            if node_outside_selection(child, self.formatting_config):
                statements.append(child)
            elif child.type in (Token.EOL, Token.COMMENT):
                statements.append(left_align(child))
            else:
                statements.append(list(tokens_by_lines(child)))
        nodes_to_be_aligned = [st for st in statements if isinstance(st, list)]
        if not nodes_to_be_aligned:
            return node
        look_up = self.create_look_up(nodes_to_be_aligned)  # for every col find longest value
        node.body = self.align_rows(statements, look_up)
        return node

    def align_rows(self, statements, look_up):
        aligned_statements = []
        for st in statements:
            if not isinstance(st, list):
                aligned_statements.append(st)
                continue
            keyword_statement = st[0][0].type in self.TOKENS_WITH_KEYWORDS
            aligned_statement = []
            for line in st:
                if is_blank_multiline(line):
                    line[-1].value = line[-1].value.lstrip(" \t")  # normalize eol from '  \n' to '\n'
                    aligned_statement.extend(line)
                    continue
                keyword_arg = keyword_statement and line[0].type == Token.CONTINUATION
                up_to = self.up_to_column if self.up_to_column != -1 else len(line) - 2
                for index, token in enumerate(line[:-2]):
                    aligned_statement.append(token)
                    separator = self.calc_separator(index, up_to, keyword_arg, token, look_up)
                    aligned_statement.append(Token(Token.SEPARATOR, separator))
                last_token = line[-2]
                # remove leading whitespace before token
                last_token.value = last_token.value.strip() if last_token.value else last_token.value
                aligned_statement.append(last_token)
                aligned_statement.append(line[-1])  # eol
            aligned_statements.append(Statement.from_tokens(aligned_statement))
        return aligned_statements

    def calc_separator(self, index, up_to, keyword_arg, token, look_up):
        if index < up_to:
            if self.min_width:
                return max(self.min_width - len(token.value), self.formatting_config.space_count) * " "
            arg_indent = self.argument_indent if keyword_arg else 0
            if keyword_arg and index != 0:
                return (
                    max(
                        (look_up[index] - len(token.value) - arg_indent + 4),
                        self.formatting_config.space_count,
                    )
                    * " "
                )
            else:
                return (look_up[index] - len(token.value) + arg_indent + 4) * " "
        else:
            return self.formatting_config.space_count * " "

    def create_look_up(self, statements):
        look_up = defaultdict(int)
        for st in statements:
            is_doc = st[0][0].type == Token.DOCUMENTATION
            for line in st:
                if is_doc:
                    up_to = 1
                elif self.up_to_column != -1:
                    up_to = self.up_to_column
                else:
                    up_to = len(line)
                for index, token in enumerate(line[:up_to]):
                    look_up[index] = max(look_up[index], len(token.value))
        return {index: round_to_four(length) for index, length in look_up.items()}



from _pytest.assertion import register_assert_rewrite

register_assert_rewrite("robocorp_ls_core.unittest_tools.fixtures")
register_assert_rewrite("robocorp_ls_core.unittest_tools.language_server_client")



import json

from tests_python.debugger_unittest import ReaderThread, IS_JYTHON
import pytest
import socket
from _pydev_bundle import pydev_localhost
from _pydevd_bundle.pydevd_comm import start_client

ABORT_CONNECTION = 'ABORT_CONNECTION'

pytestmark = pytest.mark.skipif(IS_JYTHON, reason='Getting the actual port does not work when the port is == 0.')


class _DummySocket(object):

    def __init__(self):
        self._sock_for_reader_thread = None
        self._sock_for_fixture_test = None
        self._socket_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._localhost = host = pydev_localhost.get_localhost()
        self._socket_server.bind((host, 0))
        self._socket_server.listen(1)

    def recv(self, *args, **kwargs):
        if self._sock_for_reader_thread is None:
            sock, _addr = self._socket_server.accept()
            self._sock_for_reader_thread = sock
        return self._sock_for_reader_thread.recv(*args, **kwargs)

    def put(self, msg):
        if not isinstance(msg, bytes):
            msg = msg.encode('utf-8')

        if self._sock_for_fixture_test is None:
            self._sock_for_fixture_test = start_client(*self._socket_server.getsockname())

        self._sock_for_fixture_test.sendall(msg)

    def close(self):
        self._socket_server.close()

        if self._sock_for_fixture_test is not None:
            self._sock_for_fixture_test.close()

        if self._sock_for_reader_thread is not None:
            self._sock_for_reader_thread.close()


@pytest.yield_fixture
def _dummy_socket():
    sock = _DummySocket()
    yield sock
    sock.close()


def test_fixture_reader_thread1(_dummy_socket):
    sock = _dummy_socket

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    json_part = json.dumps({'key': 'val'})
    json_part = json_part.replace(':', ':\n')
    msg = json_part

    msg = ('Content-Length: %s\r\n\r\n%s' % (len(msg), msg)).encode('utf-8')
    # Check that receiving 2 messages at a time we're able to properly deal
    # with each one.
    sock.put(msg + msg)

    assert reader_thread.get_next_message('check 1') == json_part
    assert reader_thread.get_next_message('check 2') == json_part


def test_fixture_reader_thread2(_dummy_socket):
    sock = _DummySocket()

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    json_part = json.dumps({'key': 'val'})
    json_part = json_part.replace(':', ':\n')
    msg = json_part

    http = ('Content-Length: %s\r\n\r\n%s' % (len(msg), msg))
    sock.put('msg1\nmsg2\nmsg3\n' + http + http)

    assert reader_thread.get_next_message('check 1') == 'msg1'
    assert reader_thread.get_next_message('check 2') == 'msg2'
    assert reader_thread.get_next_message('check 3') == 'msg3'
    assert reader_thread.get_next_message('check 4') == json_part
    assert reader_thread.get_next_message('check 5') == json_part


def test_fixture_reader_thread3(_dummy_socket):
    sock = _DummySocket()

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    msg = 'aaaaaaabbbbbbbccccccc'
    http = ('Content-Length: %s\r\n\r\n%s' % (len(msg), msg))
    http *= 2
    initial = http
    for i in range(1, len(http)):
        while http:
            sock.put(http[:i])
            http = http[i:]

        assert reader_thread.get_next_message('check 1: %s' % i) == msg
        assert reader_thread.get_next_message('check 2: %s' % i) == msg
        http = initial




from concurrent import futures
from functools import partial
import itertools
from typing import Dict, Union

from robocorp_ls_core.protocols import (
    IConfig,
    ActionResultDict,
    ActionResult,
    Sentinel,
    IEndPoint,
    IFuture,
)
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.pluginmanager import PluginManager

log = get_logger(__name__)


class _RfInfo:
    def __init__(
        self,
        rf_interpreter_server_manager,
        commands_thread_pool: futures.ThreadPoolExecutor,
        ls_thread_pool: futures.ThreadPoolExecutor,
    ):
        from robotframework_interactive.server.rf_interpreter_server_manager import (
            RfInterpreterServerManager,
        )

        self.interpreter: RfInterpreterServerManager = rf_interpreter_server_manager
        self.commands_thread_pool = commands_thread_pool
        self.ls_thread_pool = ls_thread_pool


class _RfInterpretersManager:
    def __init__(self, endpoint: IEndPoint, pm: PluginManager):
        self._interpreter_id_to_rf_info: Dict[int, _RfInfo] = {}
        self._next_interpreter_id = partial(next, itertools.count(0))
        self._endpoint = endpoint
        self._pm = pm

    def interpreter_start(
        self, arguments, config: IConfig
    ) -> Union[IFuture[ActionResultDict], ActionResultDict]:

        if not arguments:
            return ActionResult(
                False, message="Expected arguments ([{'uri': <uri>}])"
            ).as_dict()
        if not isinstance(arguments, (list, tuple)):
            return ActionResult(
                False, message=f"Arguments should be a Tuple[Dict]. Found: {arguments}"
            ).as_dict()

        args: dict = arguments[0]
        uri = args.get("uri", Sentinel.SENTINEL)
        if uri is Sentinel.SENTINEL:
            return ActionResult(
                False, message=f"Did not find 'uri' in {args}"
            ).as_dict()

        if not uri.endswith(".robot"):
            # We can't leave the URI as a .resource because then the parsing
            # we'd use is the one related to .resources (which can't have
            # tests) and then we wouldn't be able to create our current test
            # to be started in the interactive console.
            import os

            uri = os.path.splitext(uri)[0] + ".robot"

        # Thread pool with 1 worker (so, we're mostly sequentializing the
        # work to be done to another thread).
        commands_thread_pool = futures.ThreadPoolExecutor(max_workers=1)

        def run():
            from robotframework_ls import import_rf_interactive
            from robotframework_ls.config_extension import (
                apply_interpreter_info_to_config,
            )
            from robocorp_ls_core.ep_resolve_interpreter import EPResolveInterpreter
            from robocorp_ls_core import uris
            from robotframework_ls.impl.robot_lsp_constants import (
                OPTION_ROBOT_PYTHON_ENV,
            )

            import_rf_interactive()

            from robocorp_ls_core.options import Setup

            try:
                from robotframework_interactive.server.rf_interpreter_server_manager import (
                    RfInterpreterServerManager,
                )

                interpreter_id = self._next_interpreter_id()

                def on_interpreter_message(interpreter_message: dict):
                    """
                    :param interpreter_message:
                    Something as:
                    {
                        "jsonrpc": "2.0",
                        "method": "interpreter/output",
                        "params": {
                            "output": "Some output\n",
                            "category": "stdout",
                        },
                    }
                    """

                    params = interpreter_message["params"]
                    params["interpreter_id"] = interpreter_id
                    self._endpoint.notify(interpreter_message["method"], params)

                rf_interpreter_server_manager = RfInterpreterServerManager(
                    verbose=Setup.options.verbose,
                    base_log_file=Setup.options.log_file,
                    on_interpreter_message=on_interpreter_message,
                    uri=uri,
                )
                fs_path = uris.to_fs_path(uri)
                rf_interpreter_server_manager.config = config
                rf_config = rf_interpreter_server_manager.config

                # Just making sure that it has its own private copy before
                # mutating it...
                assert rf_config is not config

                info = {}
                for ep in self._pm.get_implementations(EPResolveInterpreter):
                    interpreter_info = ep.get_interpreter_info_for_doc_uri(uri)
                    if interpreter_info is not None:
                        target = str(interpreter_info.get_interpreter_id())
                        info["target"] = target
                        apply_interpreter_info_to_config(rf_config, interpreter_info)

                        existing_env = rf_config.get_setting(
                            OPTION_ROBOT_PYTHON_ENV, dict, {}
                        )

                        # Now, verify whether we have a
                        # 'robocorp.updateLaunchEnv', which is an additional
                        # hook that can be called to update the environment for
                        # launches (only available when we do have a custom
                        # interpreter set for it).
                        command_future = self._endpoint.request(
                            "$/executeWorkspaceCommand",
                            {
                                "command": "robocorp.updateLaunchEnv",
                                "arguments": {
                                    "targetRobot": target,
                                    "env": existing_env,
                                },
                            },
                        )
                        try:
                            new_env = command_future.result()
                            if new_env == "cancelled":
                                return ActionResult(
                                    False, message="Launch cancelled"
                                ).as_dict()
                        except:
                            log.exception(
                                "Unable to execute workspace command from the extension."
                            )
                        else:
                            if new_env:
                                if isinstance(new_env, dict):
                                    rf_config.update_override_settings(
                                        {OPTION_ROBOT_PYTHON_ENV: new_env}
                                    )
                                else:
                                    log.critical(
                                        "Expected robocorp.updateLaunchEnv to return a dict. Returned: %s (%s)",
                                        new_env,
                                        type(new_env),
                                    )

                        break

                info["path"] = str(fs_path)
                import json

                on_interpreter_message(
                    {
                        "jsonrpc": "2.0",
                        "method": "interpreter/output",
                        "params": {"output": json.dumps(info), "category": "json_info"},
                    }
                )
                rf_interpreter_server_manager.interpreter_start(uri)
                ls_thread_pool = futures.ThreadPoolExecutor(max_workers=2)
                self._interpreter_id_to_rf_info[interpreter_id] = _RfInfo(
                    rf_interpreter_server_manager, commands_thread_pool, ls_thread_pool
                )

            except Exception as e:
                log.exception("Error starting interpreter.")
                return ActionResult(False, message=str(e)).as_dict()
            else:
                return ActionResult(
                    True, result={"interpreter_id": interpreter_id}
                ).as_dict()

        return commands_thread_pool.submit(run)

    def get_interpreter_from_arguments(
        self, arguments
    ) -> Union[_RfInfo, ActionResultDict]:
        from robotframework_ls import import_rf_interactive

        import_rf_interactive()

        if not arguments:
            return ActionResult(
                False, message="Expected arguments ([{'interpreter_id': <id>}])"
            ).as_dict()
        if not isinstance(arguments, (list, tuple)):
            return ActionResult(
                False, message=f"Arguments should be a Tuple[Dict]. Found: {arguments}"
            ).as_dict()

        args: dict = arguments[0]
        interpreter_id = args.get("interpreter_id", Sentinel.SENTINEL)
        if interpreter_id is Sentinel.SENTINEL:
            return ActionResult(
                False, message=f"Did not find 'interpreter_id' in {args}"
            ).as_dict()

        rf_info = self._interpreter_id_to_rf_info.get(interpreter_id)
        if rf_info is None:
            return ActionResult(
                False, message=f"Did not find interpreter with id: {interpreter_id}"
            ).as_dict()
        return rf_info

    def interpreter_evaluate(
        self, arguments
    ) -> Union[IFuture[ActionResultDict], ActionResultDict]:
        from robotframework_ls import import_rf_interactive

        import_rf_interactive()

        from robotframework_interactive.server.rf_interpreter_server_manager import (
            RfInterpreterServerManager,
        )

        rf_info_or_dict_error: Union[
            _RfInfo, ActionResultDict
        ] = self.get_interpreter_from_arguments(arguments)
        if isinstance(rf_info_or_dict_error, dict):
            return rf_info_or_dict_error

        interpreter: RfInterpreterServerManager = rf_info_or_dict_error.interpreter
        args: dict = arguments[0]
        code = args.get("code", Sentinel.SENTINEL)
        if code is Sentinel.SENTINEL:
            return ActionResult(
                False, message=f"Did not find 'code' in {args}"
            ).as_dict()

        if interpreter.waiting_input:
            return interpreter.interpreter_evaluate(code)

        else:

            def run():
                return interpreter.interpreter_evaluate(code)

            return rf_info_or_dict_error.commands_thread_pool.submit(run)

    def interpreter_stop(
        self, arguments
    ) -> Union[IFuture[ActionResultDict], ActionResultDict]:
        from robotframework_ls import import_rf_interactive

        import_rf_interactive()

        rf_info_or_dict_error: Union[
            _RfInfo, ActionResultDict
        ] = self.get_interpreter_from_arguments(arguments)
        if isinstance(rf_info_or_dict_error, dict):
            return rf_info_or_dict_error

        rf_info: _RfInfo = rf_info_or_dict_error

        def run():
            try:
                return rf_info.interpreter.interpreter_stop()
            finally:
                try:
                    rf_info.commands_thread_pool.shutdown(wait=False)
                except:
                    log.exception("Error shutting down commands thread pool.")
                try:
                    rf_info.ls_thread_pool.shutdown(wait=False)
                except:
                    log.exception("Error shutting down ls thread pool.")

        return rf_info_or_dict_error.commands_thread_pool.submit(run)


def _handle_semantic_tokens(
    language_server_impl, rf_interpreters_manager: _RfInterpretersManager, arguments
):
    # When the user is entering text in the interpreter, the text
    # may not be the full text or it may be based on the text previously
    # entered, so, we need to ask the interpreter to compute the full
    # text so that we can get the semantic tokens based on the full
    # text that'll actually be evaluated.

    rf_info_or_dict_error: Union[
        _RfInfo, ActionResultDict
    ] = rf_interpreters_manager.get_interpreter_from_arguments(arguments)
    if isinstance(rf_info_or_dict_error, dict):
        msg = rf_info_or_dict_error.get("message")
        if msg:
            log.info(msg)
        return {"resultId": None, "data": []}

    from robotframework_interactive.server.rf_interpreter_server_manager import (
        RfInterpreterServerManager,
    )

    interpreter: RfInterpreterServerManager = rf_info_or_dict_error.interpreter
    uri = interpreter.uri

    api = language_server_impl._server_manager.get_others_api_client(uri)
    if api is None:
        log.info(
            "Unable to get api client when computing semantic tokens (for interactive usage)."
        )
        return {"resultId": None, "data": []}

    if not arguments or not isinstance(arguments, (list, tuple)) or len(arguments) != 1:
        log.info(f"Expected arguments to be a list of size 1. Found: {arguments}")
        return {"resultId": None, "data": []}

    def run():
        try:
            args: dict = arguments[0]
            code = args.get("code", Sentinel.SENTINEL)
            if code is Sentinel.SENTINEL:
                log.info(f"Did not find 'code' in {args}")
                return {"resultId": None, "data": []}

            evaluate_text_result = (
                rf_info_or_dict_error.interpreter.interpreter_compute_evaluate_text(
                    code
                )
            )
            if not evaluate_text_result["success"]:
                log.info(
                    "Unable to get code to evaluate semantic tokens (for interactive usage)."
                )
                return {"resultId": None, "data": []}
            else:
                code = evaluate_text_result["result"]

                return language_server_impl._threaded_api_request_no_doc(
                    api,
                    "request_semantic_tokens_from_code_full",
                    prefix=code["prefix"],
                    full_code=code["full_code"],
                    indent=code["indent"],
                    uri=uri,
                    monitor=None,
                )
        except:
            log.exception(f"Error computing semantic tokens for arguments: {arguments}")
            return {"resultId": None, "data": []}

    return rf_info_or_dict_error.ls_thread_pool.submit(run)


def _handle_completions(language_server_impl, rf_interpreters_manager, arguments):
    rf_info_or_dict_error: Union[
        _RfInfo, ActionResultDict
    ] = rf_interpreters_manager.get_interpreter_from_arguments(arguments)
    if isinstance(rf_info_or_dict_error, dict):
        msg = rf_info_or_dict_error.get("message")
        if msg:
            log.info(msg)
        return {"suggestions": []}

    from robotframework_interactive.server.rf_interpreter_server_manager import (
        RfInterpreterServerManager,
    )

    interpreter: RfInterpreterServerManager = rf_info_or_dict_error.interpreter
    uri = interpreter.uri

    api = language_server_impl._server_manager.get_regular_rf_api_client(uri)
    if api is None:
        log.info(
            "Unable to get api client when computing completions (for interactive usage)."
        )
        return {"suggestions": []}

    if not arguments or not isinstance(arguments, (list, tuple)) or len(arguments) != 1:
        log.info(f"Expected arguments to be a list of size 1. Found: {arguments}")
        return {"suggestions": []}

    def run():

        try:
            args: dict = arguments[0]
            code = args.get("code", Sentinel.SENTINEL)
            if code is Sentinel.SENTINEL:
                log.info(f"Did not find 'code' in {args}")
                return {"suggestions": []}

            position = args.get("position", Sentinel.SENTINEL)
            if position is Sentinel.SENTINEL:
                log.info(f"Did not find 'position' in {args}")
                return {"suggestions": []}

            # context = args.get("context", Sentinel.SENTINEL)
            # if context is Sentinel.SENTINEL:
            #     pass

            evaluate_text_result = interpreter.interpreter_compute_evaluate_text(
                code, "completions"
            )
            if not evaluate_text_result["success"]:
                log.info(
                    "Unable to get code to evaluate completions (for interactive usage)."
                )
                return {"suggestions": []}
            else:
                code = evaluate_text_result["result"]

                return language_server_impl._threaded_api_request_no_doc(
                    api,
                    "request_monaco_completions_from_code",
                    prefix=code["prefix"],
                    full_code=code["full_code"],
                    indent=code["indent"],
                    uri=uri,
                    position=position,
                    monitor=None,
                )
        except:
            log.exception(f"Error computing completions for arguments: {arguments}")
            return {"suggestions": []}

    return rf_info_or_dict_error.ls_thread_pool.submit(run)


def _handle_resolve_completion(
    language_server_impl, rf_interpreters_manager, arguments
):
    rf_info_or_dict_error: Union[
        _RfInfo, ActionResultDict
    ] = rf_interpreters_manager.get_interpreter_from_arguments(arguments)
    if isinstance(rf_info_or_dict_error, dict):
        msg = rf_info_or_dict_error.get("message")
        if msg:
            log.info(msg)
        return None

    from robotframework_interactive.server.rf_interpreter_server_manager import (
        RfInterpreterServerManager,
    )

    interpreter: RfInterpreterServerManager = rf_info_or_dict_error.interpreter
    uri = interpreter.uri

    api = language_server_impl._server_manager.get_regular_rf_api_client(uri)
    if api is None:
        log.info(
            "Unable to get api client when resolving completion (for interactive usage)."
        )
        return None

    if not arguments or not isinstance(arguments, (list, tuple)) or len(arguments) != 1:
        log.info(f"Expected arguments to be a list of size 1. Found: {arguments}")
        return None

    def run():

        try:
            args: dict = arguments[0]
            completion_item = args.get("completionItem", Sentinel.SENTINEL)
            if completion_item is Sentinel.SENTINEL:
                log.info(f"Did not find 'completionItem' in {args}")
                return None

            return language_server_impl._threaded_api_request_no_doc(
                api,
                "request_monaco_resolve_completion",
                completion_item=completion_item,
                monitor=None,
            )
        except:
            log.exception(f"Error computing completions for arguments: {arguments}")
            return {"suggestions": []}

    return rf_info_or_dict_error.ls_thread_pool.submit(run)


def execute_command(
    command,
    language_server_impl,
    rf_interpreters_manager: _RfInterpretersManager,
    arguments,
):
    from robotframework_ls.commands import (
        ROBOT_INTERNAL_RFINTERACTIVE_START,
        ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE,
        ROBOT_INTERNAL_RFINTERACTIVE_STOP,
        ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS,
        ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS,
        ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION,
    )

    if command == ROBOT_INTERNAL_RFINTERACTIVE_START:
        return rf_interpreters_manager.interpreter_start(
            arguments, language_server_impl.config
        )

    elif command == ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE:
        return rf_interpreters_manager.interpreter_evaluate(arguments)

    elif command == ROBOT_INTERNAL_RFINTERACTIVE_STOP:
        return rf_interpreters_manager.interpreter_stop(arguments)

    elif command == ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS:
        return _handle_semantic_tokens(
            language_server_impl, rf_interpreters_manager, arguments
        )

    elif command == ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS:
        return _handle_completions(
            language_server_impl, rf_interpreters_manager, arguments
        )

    elif command == ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION:
        return _handle_resolve_completion(
            language_server_impl, rf_interpreters_manager, arguments
        )



"""
Every issue is reported as ``robocop.rules.Message`` object. It can be later printed or used by
post-run reports.

Output message format
---------------------

Output message of rules can be defined with ``-f`` / ``--format`` argument. Default value::

    "{source}:{line}:{col} [{severity}] {rule_id} {desc} ({name})"

.. dropdown:: Available formats:

    * ``source``:     path to the file where the issue occurred
    * ``source_rel``: path to the file where the issue occurred, relative to execution directory
    * ``line``:       line number where the issue starts
    * ``end_line``:   line number where the issue ends
    * ``col``:        column number where the issue starts
    * ``end_col``:    column number where the issue ends
    * ``severity``:   severity of the issue, value of ``robocop.rules.RuleSeverity`` enum
    * ``rule_id``:    rule id (e.g. ``0501``)
    * ``name``:       rule name (e.g. ``line-too-long``)
    * ``desc``:       description of the rule

"""
from enum import Enum
from textwrap import dedent
from functools import total_ordering
from typing import Any, Callable, Union, Pattern, Dict, Optional
from packaging.specifiers import SpecifierSet

from jinja2 import Template

import robocop.exceptions
from robocop.utils import ROBOT_VERSION


@total_ordering
class RuleSeverity(Enum):
    """
    Rule severity.
    It can be configured with ``--configure id_or_msg_name:severity:value``
    where value can be first letter of severity value or whole name, case-insensitive.
    For example ::

        -c line-too-long:severity:e

    will change `line-too-long` rule severity to error.

    You can filter out all rules below given severity value by using following option::

        -t/--threshold <severity value>

    To only report rules with severity W and above::

        --threshold W

    """

    INFO = "I"
    WARNING = "W"
    ERROR = "E"

    @classmethod
    def parser(cls, value: Union[str, "RuleSeverity"]) -> "RuleSeverity":
        # parser can be invoked from Rule() with severity=RuleSeverity.WARNING (enum directly) or
        # from configuration with severity:W (string representation)
        severity = {
            "error": cls.ERROR,
            "e": cls.ERROR,
            "warning": cls.WARNING,
            "w": cls.WARNING,
            "info": cls.INFO,
            "i": cls.INFO,
        }.get(str(value).lower(), None)
        if severity is None:
            raise ValueError(f"Chose one of: {', '.join(sev.value for sev in cls)}") from None
        return severity

    def __str__(self):
        return self.value

    def __lt__(self, other):
        look_up = [sev.value for sev in RuleSeverity]
        return look_up.index(self.value) < look_up.index(other.value)

    def diag_severity(self) -> int:
        return {"I": 3, "W": 2, "E": 1}.get(self.value, 4)


class RuleParam:
    """
    Parameter of the Rule.
    Each rule can have number of parameters (default one is severity).
    """

    def __init__(self, name: str, default: Any, converter: Callable, desc: str):
        """
        :param name: Name of the parameter used when configuring rule (also displayed in the docs)
        :param default: Default value of the parameter
        :param converter: Method used for converting from string. It can be separate method or classmethod from
        particular class (see `:RuleSeverity:` for example of class that is used as rule parameter value).
        It must return value
        :param desc: Description of rule parameter
        """
        self.name = name
        self.converter = converter
        self.desc = desc
        self.raw_value = None
        self._value = None
        self.value = default

    def __str__(self):
        s = f"{self.name} = {self.raw_value}\n" f"        type: {self.converter.__name__}"
        if self.desc:
            s += "\n" f"        info: {self.desc}"
        return s

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self.raw_value = value  # useful for docs/printing
        try:
            self._value = self.converter(value)
        except ValueError as err:
            raise robocop.exceptions.RuleParamFailedInitError(self, value, str(err)) from None


class Rule:
    """
    Robocop linter rule.
    It can be used for reporting issues that are breaking particular rule.
    You can store configuration of the rule inside RuleParam parameters.

    Every rule contains one default RuleParam - severity.
    """

    def __init__(
        self,
        *params: RuleParam,
        rule_id: str,
        name: str,
        msg: str,
        severity: RuleSeverity,
        version: str = None,
        docs: str = "",
    ):
        """
        :param params: RuleParam() instances
        :param rule_id: id of the rule
        :param name: name of the rule
        :param msg: message printed when rule breach is detected
        :param severity: severity of the rule (ie: RuleSeverity.INFO)
        :param version: supported Robot Framework version (ie: >=4.0)
        :param docs: Full documentation of the rule (rst supported)
        description of the rule
        """
        self.rule_id = rule_id
        self.name = name
        self.msg = msg
        self.msg_template = self.get_template(msg)
        self.docs = dedent(docs)
        self.config = {
            "severity": RuleParam(
                "severity", severity, RuleSeverity.parser, "Rule severity (E = Error, W = Warning, I = Info)"
            )
        }
        for param in params:
            self.config[param.name] = param
        self.enabled = True
        self.supported_version = version if version else "All"
        self.enabled_in_version = self.supported_in_rf_version(version)

    @property
    def severity(self):
        return self.config["severity"].value

    @staticmethod
    def supported_in_rf_version(version: str) -> bool:
        if not version:
            return True
        return ROBOT_VERSION in SpecifierSet(version, prereleases=True)

    @staticmethod
    def get_template(msg: str) -> Optional[Template]:
        if "{" in msg:
            return Template(msg)
        return None

    def get_message(self, **kwargs):
        if self.msg_template:
            return self.msg_template.render(**kwargs)
        return self.msg

    def __str__(self):
        return (
            f"Rule - {self.rule_id} [{self.config['severity'].value}]: {self.name}: {self.msg} "
            f"({self.get_enabled_status_desc()})"
        )

    def get_enabled_status_desc(self):
        s = "enabled" if self.enabled else "disabled"
        if not self.enabled and self.supported_version != "All":
            s += f" - supported only for RF version {self.supported_version}"
        return s

    def configure(self, param, value):
        if param not in self.config:
            raise robocop.exceptions.ConfigGeneralError(
                f"Provided param '{param}' for rule '{self.name}' does not exist. "
                f"Available configurable(s) for this rule:\n"
                f"    {self.available_configurables()}"
            )
        self.config[param].value = value

    def available_configurables(self, include_severity: bool = True):
        params = [str(param) for param in self.config.values() if param.name != "severity" or include_severity]
        if not params:
            return ""
        return "\n    ".join(params)

    def prepare_message(self, source, node, lineno, col, end_lineno, end_col, ext_disablers, **kwargs):
        msg = self.get_message(**kwargs)
        return Message(
            rule=self,
            msg=msg,
            source=source,
            node=node,
            lineno=lineno,
            col=col,
            end_col=end_col,
            end_lineno=end_lineno,
            ext_disablers=ext_disablers,
        )

    def matches_pattern(self, pattern: Union[str, Pattern]):
        """check if this rule matches given pattern"""
        if isinstance(pattern, str):
            return pattern in (self.name, self.rule_id)
        return pattern.match(self.name) or pattern.match(self.rule_id)


class Message:
    def __init__(
        self,
        rule: Rule,
        msg,
        source,
        node,
        lineno,
        col,
        end_lineno,
        end_col,
        ext_disablers=None,
    ):
        self.enabled = rule.enabled
        self.rule_id = rule.rule_id
        self.name = rule.name
        self.severity = rule.severity
        self.desc = msg
        self.source = source
        self.line = 1
        if node is not None and node.lineno > -1:
            self.line = node.lineno
        if lineno is not None:
            self.line = lineno
        self.col = 1 if col is None else col
        self.end_line = self.line if end_lineno is None else end_lineno
        self.end_col = self.col if end_col is None else end_col
        self.ext_disablers = ext_disablers if ext_disablers else []

    def __lt__(self, other):
        return (self.line, self.col, self.rule_id) < (
            other.line,
            other.col,
            other.rule_id,
        )

    def get_fullname(self) -> str:
        return f"{self.severity.value}{self.rule_id} ({self.name})"

    def to_json(self) -> Dict:
        return {
            "source": self.source,
            "line": self.line,
            "column": self.col,
            "severity": self.severity.value,
            "rule_id": self.rule_id,
            "description": self.desc,
            "rule_name": self.name,
        }



from typing import Dict, Optional
import re

from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.ast_utils import MAX_ERRORS
from robotframework_ls.impl.protocols import (
    IKeywordFound,
    IKeywordCollector,
    ICompletionContext,
    ILibraryDoc,
    INode,
    IVariableFound,
    AbstractVariablesCollector,
)
from robocorp_ls_core.lsp import DiagnosticSeverity, DiagnosticTag
from robotframework_ls.impl.robot_lsp_constants import (
    OPTION_ROBOT_LINT_VARIABLES,
    OPTION_ROBOT_LINT_IGNORE_VARIABLES,
)
from functools import lru_cache


log = get_logger(__name__)


class _KeywordContainer(object):
    def __init__(self) -> None:
        self._name_to_keyword: Dict[str, IKeywordFound] = {}
        self._names_with_variables: Dict[str, IKeywordFound] = {}

    def add_keyword(self, keyword_found: IKeywordFound) -> None:
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        normalized_name = normalize_robot_name(keyword_found.keyword_name)
        self._name_to_keyword[normalized_name] = keyword_found

        if "{" in normalized_name:
            self._names_with_variables[normalized_name] = keyword_found

    def get_keyword(self, normalized_keyword_name: str) -> Optional[IKeywordFound]:
        from robotframework_ls.impl.text_utilities import matches_name_with_variables

        keyword_found = self._name_to_keyword.get(normalized_keyword_name)

        if keyword_found is not None:
            return keyword_found

        # We do not have an exact match, still, we need to check if we may
        # have a match in keywords that accept variables.
        for name, keyword_found in self._names_with_variables.items():
            if matches_name_with_variables(normalized_keyword_name, name):
                return keyword_found

        return None


class _VariablesCollector(AbstractVariablesCollector):
    def __init__(self, on_unresolved_variable_import):
        self._variables_collected = set()
        self._template_variables_collected = []
        self.on_unresolved_variable_import = on_unresolved_variable_import

    def accepts(self, variable_name: str) -> bool:
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        normalized = normalize_robot_name(variable_name)
        if "{" in normalized:
            self._template_variables_collected.append(normalized)
        self._variables_collected.add(normalized)
        # We don't want to create the IVariableFound, just the names should be
        # enough for our usage.
        return False

    def on_variable(self, variable_found: IVariableFound):
        pass

    def contains_variable(self, variable_name):
        from robotframework_ls.impl.text_utilities import matches_name_with_variables

        if variable_name in self._variables_collected:
            return True

        for template_var in self._template_variables_collected:
            if matches_name_with_variables(variable_name, template_var):
                return True

        return False


class _AnalysisKeywordsCollector(object):
    def __init__(
        self,
        on_unresolved_library,
        on_unresolved_resource,
        on_resolved_library,
    ):
        self._keywords_container = _KeywordContainer()
        self._resource_name_to_keywords_container = {}
        self._library_name_to_keywords_container = {}

        self.on_unresolved_library = on_unresolved_library
        self.on_unresolved_resource = on_unresolved_resource
        self.on_resolved_library = on_resolved_library

    def accepts(self, keyword_name):
        return True

    def on_keyword(self, keyword_found: IKeywordFound):
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        # Note: Even if something is imported 'WITH NAME', it's still added
        # to the global scope (there's just an additional reference with the
        # new name).
        self._keywords_container.add_keyword(keyword_found)
        library_name = keyword_found.library_name
        library_alias = keyword_found.library_alias
        resource_name = keyword_found.resource_name

        if library_name:
            if library_alias:
                name = normalize_robot_name(library_alias)
            else:
                name = normalize_robot_name(library_name)
            dct = self._library_name_to_keywords_container
        elif resource_name:
            name = normalize_robot_name(resource_name)
            dct = self._resource_name_to_keywords_container
        else:
            log.info(
                "No library name nor resource name for keyword: %s",
                keyword_found.keyword_name,
            )
            return

        keyword_container = dct.get(name)
        if keyword_container is None:
            keyword_container = dct[name] = _KeywordContainer()

        keyword_container.add_keyword(keyword_found)

    def get_keyword(self, normalized_keyword_name: str) -> Optional[IKeywordFound]:
        from robotframework_ls.impl import text_utilities

        keyword_found = self._keywords_container.get_keyword(normalized_keyword_name)
        if keyword_found is not None:
            return keyword_found

        # Note: the name could be something as `alias.keywordname` or
        # 'libraryname.keywordname`. In this case, we need to verify if there's
        # a library/alias with that specific name.
        for name, remainder in text_utilities.iter_dotted_names(
            normalized_keyword_name
        ):
            if not name or not remainder:
                continue
            containers = []
            keywords_container = self._resource_name_to_keywords_container.get(name)
            if keywords_container:
                containers.append(keywords_container)
            keywords_container = self._library_name_to_keywords_container.get(name)
            if keywords_container:
                containers.append(keywords_container)

            for keywords_container in containers:
                keyword_found = keywords_container.get_keyword(remainder)
                if keyword_found is not None:
                    return keyword_found

        return None

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


def collect_analysis_errors(initial_completion_context):
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.ast_utils import create_error_from_node
    from robotframework_ls.impl.collect_keywords import collect_keywords
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.text_utilities import contains_variable_text
    from robotframework_ls.impl.keyword_argument_analysis import (
        UsageInfoForKeywordArgumentAnalysis,
    )
    from robot.api import Token

    errors = []
    config = initial_completion_context.config

    def on_resolved_library(
        completion_context: ICompletionContext,
        library_node: Optional[INode],
        library_doc: ILibraryDoc,
    ):
        if library_node is None:
            return

        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS, bool, True
        ):
            return

        doc = completion_context.doc
        if not doc or doc.uri != initial_completion_context.doc.uri:
            return

        from robotframework_ls.impl.keyword_argument_analysis import (
            KeywordArgumentAnalysis,
        )

        library_args = []
        if library_doc.inits:
            keyword_doc = library_doc.inits[0]
            library_args = keyword_doc.args

        # Ok, we found the keyword, let's check if the arguments are correct.
        keyword_argument_analysis = KeywordArgumentAnalysis(library_args)

        name_token = library_node.get_token(Token.NAME)
        if name_token is not None:
            for error in keyword_argument_analysis.collect_keyword_usage_errors(
                UsageInfoForKeywordArgumentAnalysis(library_node, name_token)
            ):
                errors.append(error)

    def on_unresolved_library(
        completion_context: ICompletionContext,
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES,
        )
        from robotframework_ls.impl.robot_version import get_robot_major_version

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)

            additional = ""
            if error_msg is None:
                error_msg = ""
            else:
                if "expected" in error_msg:
                    if re.search(r"expected\s+(.*)\s+argument(s)?,\s+got", error_msg):
                        additional = f'\nConsider using default arguments in the {library_name} constructor and\ncalling the "Robot Framework: Clear caches and restart" action or\nadding "{library_name}" to the "robot.libraries.libdoc.needsArgs"\nsetting to pass the typed arguments when generating the libspec.'

                if not additional and "Importing" in error_msg:
                    if get_robot_major_version() <= 3:
                        pattern = r"Importing\s+test\s+library(.*)failed"
                    else:
                        pattern = r"Importing\s+library(.*)failed"
                    if re.search(pattern, error_msg):
                        additional = f'\nConsider adding the needed paths to the "robot.pythonpath" setting\nand calling the "Robot Framework: Clear caches and restart" action.'

            errors.append(
                ast_utils.Error(
                    f"Unresolved library: {library_name}.{error_msg}{additional}".strip(),
                    start,
                    end,
                )
            )

    def on_unresolved_resource(
        completion_context: ICompletionContext,
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_RESOURCES,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_RESOURCES, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)
            if not error_msg:
                error_msg = f"Unresolved resource: {library_name}"

            errors.append(ast_utils.Error(error_msg, start, end))

    collector = _AnalysisKeywordsCollector(
        on_unresolved_library,
        on_unresolved_resource,
        on_resolved_library,
    )
    collect_keywords(initial_completion_context, collector)

    ast = initial_completion_context.get_ast()
    for keyword_usage_info in ast_utils.iter_keyword_usage_tokens(
        ast, collect_args_as_keywords=True
    ):
        initial_completion_context.check_cancelled()
        if contains_variable_text(keyword_usage_info.name):
            continue
        normalized_name = normalize_robot_name(keyword_usage_info.name)
        keyword_found = collector.get_keyword(normalized_name)
        try:
            if not keyword_found:
                from robotframework_ls.impl.robot_lsp_constants import (
                    OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS,
                )

                if config is not None and not config.get_setting(
                    OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS, bool, True
                ):
                    continue

                node = keyword_usage_info.node
                error = create_error_from_node(
                    node,
                    "Undefined keyword: %s." % (keyword_usage_info.name,),
                    tokens=[keyword_usage_info.token],
                )
                errors.append(error)

            else:
                from robotframework_ls.impl.robot_lsp_constants import (
                    OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
                )

                if config is not None and not config.get_setting(
                    OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS, bool, True
                ):
                    continue

                from robotframework_ls.impl.keyword_argument_analysis import (
                    KeywordArgumentAnalysis,
                )

                # Ok, we found the keyword, let's check if the arguments are correct.
                keyword_argument_analysis = KeywordArgumentAnalysis(
                    keyword_found.keyword_args
                )

                keyword_token = None
                if keyword_token is None:
                    keyword_token = keyword_usage_info.node.get_token(Token.KEYWORD)

                if keyword_token is not None:
                    for error in keyword_argument_analysis.collect_keyword_usage_errors(
                        UsageInfoForKeywordArgumentAnalysis(
                            keyword_usage_info.node,
                            keyword_token,
                        )
                    ):
                        errors.append(error)

                if keyword_found.is_deprecated():
                    error = create_error_from_node(
                        keyword_usage_info.node,
                        f"Keyword: {keyword_usage_info.name} is deprecated",
                        tokens=[keyword_usage_info.token],
                    )
                    error.severity = DiagnosticSeverity.Hint
                    error.tags = [DiagnosticTag.Deprecated]
                    errors.append(error)

            if len(errors) >= MAX_ERRORS:
                # i.e.: Collect at most 100 errors
                break
        except:
            log.exception("Error collecting exceptions")

    for error in _collect_undefined_variables_errors(initial_completion_context):
        errors.append(error)
        if len(errors) >= MAX_ERRORS:
            # i.e.: Collect at most 100 errors
            break

    return errors


def _is_number_var(normalized_variable_name):
    # see: robot.variables.finders.NumberFinder
    try:
        bases = {"0b": 2, "0o": 8, "0x": 16}
        if normalized_variable_name.startswith(tuple(bases)):
            return int(
                normalized_variable_name[2:], bases[normalized_variable_name[:2]]
            )
        int(normalized_variable_name)
        return True
    except:
        pass  # Let's try float...

    try:
        float(normalized_variable_name)
        return True
    except:
        pass

    return False


def _is_python_eval_var(normalized_variable_name):
    return (
        len(normalized_variable_name) >= 2
        and normalized_variable_name[0] == "{"
        and normalized_variable_name[-1] == "}"
    )


@lru_cache(maxsize=1000)
def _skip_variable_analysis(normalized_variable_name):

    if _is_number_var(normalized_variable_name):
        return True

    if _is_python_eval_var(normalized_variable_name):
        return True

    return False


_match_extended = re.compile(
    r"""
    (.+?)          # base name (group 1)
    ([^\s\w].+)    # extended part (group 2)
""",
    re.UNICODE | re.VERBOSE,
).match


def _extract_base_from_extended_var_name(normalized_variable_name):
    m = _match_extended(normalized_variable_name)
    if m is None:
        return normalized_variable_name

    base_name, _extended = m.groups()
    return base_name


def _env_vars_upper():
    import os

    return tuple(x.upper() for x in os.environ)


def _collect_undefined_variables_errors(initial_completion_context):
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.variable_resolve import normalize_variable_name
    from robotframework_ls.impl.ast_utils import create_error_from_node

    config = initial_completion_context.config
    if config is not None and not config.get_setting(
        OPTION_ROBOT_LINT_VARIABLES, bool, True
    ):
        return

    unresolved_variable_import_errors = []

    def on_unresolved_variable_import(
        completion_context: ICompletionContext,
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS,
        )

        if config is not None and not config.get_setting(
            OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS, bool, True
        ):
            return

        doc = completion_context.doc
        if doc and doc.uri == initial_completion_context.doc.uri:
            start = (lineno - 1, col_offset)
            end = (end_lineno - 1, end_col_offset)
            if not error_msg:
                error_msg = f"Unresolved variable import: {library_name}"

            unresolved_variable_import_errors.append(
                ast_utils.Error(error_msg, start, end)
            )

    from robotframework_ls.impl import variable_completions

    ignore_variables = set()
    if config is not None:
        ignore_variables.update(
            normalize_variable_name(str(x))
            for x in config.get_setting(OPTION_ROBOT_LINT_IGNORE_VARIABLES, list, [])
        )

    ast = initial_completion_context.get_ast()

    globals_collector = _VariablesCollector(
        on_unresolved_variable_import=on_unresolved_variable_import
    )

    # Collect undefined variables
    variable_completions.collect_global_variables(
        initial_completion_context, globals_collector, only_current_doc=False
    )

    yield from iter(unresolved_variable_import_errors)

    env_vars_upper = None

    for token_info in ast_utils.iter_variable_references(ast):
        initial_completion_context.check_cancelled()

        if token_info.node.__class__.__name__ in (
            "ResourceImport",
            "LibraryImport",
            "VariableImport",
        ):
            # These ones are handled differently as it ends up in an unresolved
            # import.
            continue

        var_name = token_info.token.value
        if token_info.var_identifier == "%":
            from robotframework_ls.impl.variable_resolve import extract_variable_base

            if "=" in var_name:
                # Consider case: %{SOME_VAR=defaultvalue}
                continue

            if env_vars_upper is None:
                env_vars_upper = _env_vars_upper()

            if extract_variable_base(var_name).upper() not in env_vars_upper:
                # Environment variable
                yield create_error_from_node(
                    token_info.node,
                    f"Undefined environment variable: {token_info.token.value}",
                    tokens=[token_info.token],
                )
            continue

        normalized_variable_name = normalize_variable_name(var_name)

        found = False
        while not found:
            if normalized_variable_name in ignore_variables:
                found = True
                break

            if _skip_variable_analysis(normalized_variable_name):
                found = True
                break

            if globals_collector.contains_variable(normalized_variable_name):
                found = True
                break

            locals_collector = _VariablesCollector(lambda *args, **kwargs: None)
            local_ctx = initial_completion_context.create_copy_with_selection(
                line=token_info.token.lineno - 1,
                col=token_info.token.col_offset - 1,
            )

            variable_completions.collect_local_variables(
                local_ctx, locals_collector, token_info
            )
            if locals_collector.contains_variable(normalized_variable_name):
                found = True
                break

            extracted_base = _extract_base_from_extended_var_name(
                normalized_variable_name
            )
            if extracted_base and extracted_base != normalized_variable_name:
                normalized_variable_name = extracted_base
            else:
                break

        if not found:
            yield create_error_from_node(
                token_info.node,
                f"Undefined variable: {token_info.token.value}",
                tokens=[token_info.token],
            )



from robocorp_ls_core.config import Config

OPTION_ROBOT_PYTHON_EXECUTABLE = "robot.python.executable"

OPTION_ROBOT_PYTHON_ENV = "robot.python.env"

OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT = "robot.language-server.tcp-port"

OPTION_ROBOT_LANGUAGE_SERVER_ARGS = "robot.language-server.args"

OPTION_ROBOT_VARIABLES = "robot.variables"

OPTION_ROBOT_LINT_ROBOCOP_ENABLED = "robot.lint.robocop.enabled"

OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT = "robot.completions.keywords.format"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_FIRST_UPPER = "First upper"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_TITLE = "Title Case"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_LOWER = "all lower"
OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT_ALL_UPPER = "ALL UPPER"


OPTION_ROBOT_PYTHONPATH = "robot.pythonpath"

OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS = (
    "robot.workspaceSymbolsOnlyForOpenDocs"
)

OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM = "robot.completions.section_headers.form"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL = "plural"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_SINGULAR = "singular"
OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_BOTH = "both"

# Options which must be set as environment variables.
ENV_OPTION_ROBOT_DAP_TIMEOUT = "ROBOT_DAP_TIMEOUT"


ALL_ROBOT_OPTIONS = frozenset(
    (
        OPTION_ROBOT_PYTHON_EXECUTABLE,
        OPTION_ROBOT_PYTHON_ENV,
        OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT,
        OPTION_ROBOT_LANGUAGE_SERVER_ARGS,
        OPTION_ROBOT_VARIABLES,
        OPTION_ROBOT_LINT_ROBOCOP_ENABLED,
        OPTION_ROBOT_PYTHONPATH,
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
        OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS,
        OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT,
    )
)

# NOTE: This is a copy of the settings from the language server so that we don't
# have a cyclic dependency.


class RfInterpreterRobotConfig(Config):
    ALL_OPTIONS = ALL_ROBOT_OPTIONS



import threading
import unittest
import os
import pytest
import pydevconsole

from _pydev_bundle.pydev_imports import xmlrpclib, SimpleXMLRPCServer
from _pydev_bundle.pydev_localhost import get_localhost

try:
    raw_input
    raw_input_name = 'raw_input'
except NameError:
    raw_input_name = 'input'
    
try:
    from IPython import core  # @UnusedImport
    has_ipython = True
except:
    has_ipython = False


#=======================================================================================================================
# Test
#=======================================================================================================================
@pytest.mark.skipif(os.environ.get('TRAVIS') == 'true' or not has_ipython, reason='Too flaky on Travis (and requires IPython).')
class Test(unittest.TestCase):

    def start_client_thread(self, client_port):
        class ClientThread(threading.Thread):
            def __init__(self, client_port):
                threading.Thread.__init__(self)
                self.client_port = client_port

            def run(self):
                class HandleRequestInput:
                    def RequestInput(self):
                        client_thread.requested_input = True
                        return 'RequestInput: OK'

                    def NotifyFinished(self, *args, **kwargs):
                        client_thread.notified_finished += 1
                        return 1

                handle_request_input = HandleRequestInput()

                from _pydev_bundle import pydev_localhost
                self.client_server = client_server = SimpleXMLRPCServer((pydev_localhost.get_localhost(), self.client_port), logRequests=False)
                client_server.register_function(handle_request_input.RequestInput)
                client_server.register_function(handle_request_input.NotifyFinished)
                client_server.serve_forever()

            def shutdown(self):
                return
                self.client_server.shutdown()

        client_thread = ClientThread(client_port)
        client_thread.requested_input = False
        client_thread.notified_finished = 0
        client_thread.daemon = True
        client_thread.start()
        return client_thread


    def get_free_addresses(self):
        from _pydev_bundle.pydev_localhost import get_socket_names
        socket_names = get_socket_names(2, close=True)
        return [socket_name[1] for socket_name in socket_names]

    def test_server(self):
        # Just making sure that the singleton is created in this thread.
        from _pydev_bundle.pydev_ipython_console_011 import get_pydev_frontend
        get_pydev_frontend(get_localhost(), 0)

        client_port, server_port = self.get_free_addresses()
        class ServerThread(threading.Thread):
            def __init__(self, client_port, server_port):
                threading.Thread.__init__(self)
                self.client_port = client_port
                self.server_port = server_port

            def run(self):
                from _pydev_bundle import pydev_localhost
                print('Starting server with:', pydev_localhost.get_localhost(), self.server_port, self.client_port)
                pydevconsole.start_server(pydev_localhost.get_localhost(), self.server_port, self.client_port)
        server_thread = ServerThread(client_port, server_port)
        server_thread.daemon = True
        server_thread.start()

        client_thread = self.start_client_thread(client_port) #@UnusedVariable

        try:
            import time
            time.sleep(.3) #let's give it some time to start the threads

            from _pydev_bundle import pydev_localhost
            server = xmlrpclib.Server('http://%s:%s' % (pydev_localhost.get_localhost(), server_port))
            server.execLine("import sys; print('Running with: %s %s' % (sys.executable or sys.platform, sys.version))")
            server.execLine('class Foo:')
            server.execLine('    pass')
            server.execLine('')
            server.execLine('foo = Foo()')
            server.execLine('a = %s()' % raw_input_name)
            initial = time.time()
            while not client_thread.requested_input:
                if time.time() - initial > 2:
                    raise AssertionError('Did not get the return asked before the timeout.')
                time.sleep(.1)
            frame_xml = server.getFrame()
            self.assertTrue('RequestInput' in frame_xml, 'Did not fid RequestInput in:\n%s' % (frame_xml,))
        finally:
            client_thread.shutdown()




import subprocess
import sys

args = [
    'connect(\\"127 . 0.0.1\\")',
    '"',
    '',
    '\\',
    '\\"',
    '""',
]


def main():
    retcode = subprocess.call([sys.executable, __file__] + args)
    assert retcode == 0


if __name__ == '__main__':
    sys_args = sys.argv[1:]
    if sys_args:
        assert sys_args == args, 'Expected that %r == %r' % (sys_args, args)
        print('break here')
        print('TEST SUCEEDED!')
    else:
        main()
